import asyncio
import codecs
import json
import logging
import os
import sys
from typing import Optional

import requests
import websockets
from pydantic import BaseModel, ValidationError

from robocorp.action_server._robo_utils.process import exit_when_pid_exists

log = logging.getLogger(__name__)


class SessionPayload(BaseModel):
    sessionId: str
    sessionSecret: str


class BodyPayload(BaseModel):
    requestId: str
    path: str
    method: str = "GET"
    body: dict | None = None
    headers: dict


class ExposeSessionJson(BaseModel):
    expose_session: str
    url: str
    api_key: str | None = None


def get_expose_session_path(datadir: str) -> str:
    return os.path.join(datadir, "expose_session.json")


def read_expose_session_json(datadir: str) -> None | ExposeSessionJson:
    session_json = None
    try:
        expose_session_path = get_expose_session_path(datadir)
        log.debug(f"🗂️ Reading expose_session.json path={expose_session_path}")
        with open(expose_session_path, "r") as f:
            session_json = ExposeSessionJson(**json.load(f))
    except FileNotFoundError:
        pass
    except ValidationError as e:
        log.error("Failed to load previous expose session", e)
    except json.JSONDecodeError as e:
        log.error("Failed to decode exopse session json", e)
    return session_json


def write_expose_session_json(
    datadir: str, expose_session_json: ExposeSessionJson
) -> None:
    expose_session_path = get_expose_session_path(datadir)
    log.debug(f"🗂️ Writing expose_session.json path={expose_session_path}")
    with open(expose_session_path, "w") as f:
        json.dump(
            expose_session_json.model_dump(),
            f,
            indent=2,
        )


def get_expose_session(payload: SessionPayload) -> str:
    return f"{payload.sessionId}:{payload.sessionSecret}".encode("ascii").hex()


def get_expose_session_payload(expose_session: str) -> SessionPayload:
    session_id, session_secret = codecs.decode(
        expose_session.encode("ascii"), "hex"
    ).split(b":")
    return SessionPayload(
        sessionId=session_id.decode(), sessionSecret=session_secret.decode()
    )


def forward_request(base_url: str, payload: BodyPayload) -> requests.Response:
    url = base_url.rstrip("/") + "/" + payload.path.lstrip("/")

    if payload.method == "GET":
        return requests.get(url, headers=payload.headers)
    elif payload.method == "POST":
        return requests.post(url, json=payload.body, headers=payload.headers)
    elif payload.method == "PUT":
        return requests.put(url, json=payload.body, headers=payload.headers)
    elif payload.method == "DELETE":
        return requests.delete(url, headers=payload.headers)
    else:
        raise NotImplementedError(f"Method {payload.method} not implemented")


async def handle_ping_pong(
    ws: websockets.WebSocketClientProtocol,
    pong_queue: asyncio.Queue,
    no_connection_queue: asyncio.Queue,
):
    while True:
        await asyncio.sleep(2)
        await ws.send("ping")

        try:
            await asyncio.wait_for(pong_queue.get(), timeout=2)
        except asyncio.TimeoutError:
            log.debug("Ping-pong message timeout")
            await ws.close()
            await no_connection_queue.put("no_connection")
            break


class ExposePingTimeoutError(Exception):
    pass


async def expose_server(
    port: int,
    host: str,
    expose_url: str,
    datadir: str,
    api_key: str | None = None,
    expose_session: str | None = None,
):
    """
    Exposes the server to the world.
    """

    pong_queue: asyncio.Queue = asyncio.Queue(maxsize=1)
    no_connection_queue: asyncio.Queue = asyncio.Queue(maxsize=1)

    async def listen_for_requests() -> None:
        max_retries = 3
        retry_delay = 3
        retries = 0

        session_payload: Optional[SessionPayload] = (
            get_expose_session_payload(expose_session) if expose_session else None
        )
        while retries < max_retries:
            headers = (
                {
                    "x-session-id": session_payload.sessionId,
                    "x-session-secret": session_payload.sessionSecret,
                }
                if session_payload
                else {}
            )

            try:
                async with websockets.connect(
                    f"wss://client.{expose_url}",
                    extra_headers=headers,
                    logger=log,
                    close_timeout=0,
                ) as ws:
                    if retries > 0:
                        log.info("Reconnected to expose tunnel server")
                        retries = 0

                    ping_task = asyncio.create_task(
                        handle_ping_pong(ws, pong_queue, no_connection_queue)
                    )

                    try:
                        while True:
                            done, pending = await asyncio.wait(
                                [ws.recv(), no_connection_queue.get()],
                                return_when=asyncio.FIRST_COMPLETED,
                            )

                            for task in done:
                                message = task.result()
                                if message == "no_connection":
                                    raise ExposePingTimeoutError("No connection")

                            for task in pending:
                                task.cancel()

                            if message == "pong":
                                await pong_queue.put("pong")
                                continue

                            data = json.loads(message)

                            try:
                                session_payload = SessionPayload(**data)

                                url = (
                                    f"https://{session_payload.sessionId}.{expose_url}"
                                )
                                log.info(f"🌍 URL: {url}")
                                if api_key is not None:
                                    log.info(
                                        f'🔑 Add following header api authorization header to run actions: {{ "Authorization": "Bearer {api_key}" }}'  # noqa
                                    )
                                new_expose_session = get_expose_session(session_payload)
                                write_expose_session_json(
                                    datadir=datadir,
                                    expose_session_json=ExposeSessionJson(
                                        expose_session=new_expose_session,
                                        api_key=api_key,
                                        url=url,
                                    ),
                                )
                                continue
                            except Exception as e:
                                if not session_payload:
                                    log.error(
                                        "Unable to get session payload. Exposing the local server failed. Try again."
                                    )
                                    raise
                                pass

                            try:
                                payload = BodyPayload(**data)
                                if (
                                    payload.path != "/openapi.json"
                                    and api_key is not None
                                ):
                                    if (
                                        payload.headers.get("authorization")
                                        != f"Bearer {api_key}"
                                    ):
                                        log.error(
                                            "Request failed because the API key is invalid."
                                        )
                                        await ws.send(
                                            json.dumps(
                                                {
                                                    "requestId": payload.requestId,
                                                    "response": json.dumps(
                                                        {
                                                            "error": {
                                                                "code": "INVALID_API_KEY",
                                                                "message": "The API key is invalid.",
                                                            },
                                                        }
                                                    ),
                                                    "status": 403,
                                                }
                                            )
                                        )
                                        continue

                                base_url = f"http://{host}:{port}"
                                response: requests.Response = forward_request(
                                    base_url=base_url, payload=payload
                                )
                                await ws.send(
                                    json.dumps(
                                        {
                                            "requestId": payload.requestId,
                                            "response": json.dumps(
                                                response.json(),
                                                indent=2,
                                            ),
                                            "status": response.status_code,
                                        }
                                    )
                                )

                            except Exception as e:
                                log.error("Error forwarding request", e)
                                pass

                    except Exception as e:
                        raise e

                    finally:
                        ping_task.cancel()
                        try:
                            await ping_task
                        except asyncio.CancelledError:
                            pass

            except (
                websockets.ConnectionClosed,
                websockets.ConnectionClosedError,
                ExposePingTimeoutError,
                OSError,
            ) as e:
                retries += 1
                log.info(
                    f"Connection closed, attempting to reconnect...({retries}/{max_retries})"
                )
                log.debug(e)
                await asyncio.sleep(retry_delay * retries)

            except Exception as e:
                log.error("Error connecting to expose tunnel server", e)
                break

        else:
            log.error(
                "Could not connect to expose tunnel server. Check your connectivity and try again."
            )

    task = asyncio.create_task(listen_for_requests())
    await task  # Wait for listen_for_requests to complete


def main(parent_pid, port, verbose, host, expose_url, datadir, api_key, expose_session):
    logging.basicConfig(
        level=logging.DEBUG if verbose.count("v") > 0 else logging.INFO,
        format="%(message)s",
        datefmt="[%X]",
    )

    exit_when_pid_exists(int(parent_pid))
    try:
        asyncio.run(
            expose_server(
                port=int(port),
                host=host,
                expose_url=expose_url,
                datadir=datadir,
                api_key=api_key if api_key != "None" else None,
                expose_session=expose_session if expose_session != "None" else None,
            )
        )
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main(*sys.argv[1:])



from dataclasses import dataclass
from typing import List

from fastapi.routing import APIRouter

from robocorp.action_server._models import Action, ActionPackage

action_package_api_router = APIRouter(prefix="/api/actionPackages")


@dataclass
class ActionPackageApi:
    id: str  # primary key (uuid)
    name: str  # The name for the action package
    actions: List[Action]


@action_package_api_router.get("", response_model=List[ActionPackageApi])
def list_action_packages():
    from robocorp.action_server._models import get_db

    db = get_db()
    with db.connect():
        # We're running in the threadpool used by fast api, so, we need
        # to make a new connection (maybe it'd make sense to create a
        # connection pool instead of always creating a new connection...).
        action_packages = db.all(ActionPackage)

        id_to_action_package = {}
        for action_package in action_packages:
            id_to_action_package[action_package.id] = ActionPackageApi(
                action_package.id, action_package.name, []
            )

        for action in db.all(Action):
            id_to_action_package[action.action_package_id].actions.append(action)

    return list(id_to_action_package.values())



"""
This module contains utilities for testing and to do a 'selftest' of the 
executable even in release mode.
"""

import os
import re
import subprocess
import sys
import time
import typing
from concurrent.futures import TimeoutError
from contextlib import contextmanager
from pathlib import Path
from subprocess import CompletedProcess
from typing import Dict, Iterator, Literal, Optional, Tuple, Union

if typing.TYPE_CHECKING:
    from robocorp.action_server._robo_utils.process import Process


def is_debugger_active() -> bool:
    try:
        import pydevd  # type:ignore
    except ImportError:
        return False

    return bool(pydevd.get_global_debugger())


class ActionServerProcess:
    SHOW_OUTPUT = True

    def __init__(self, datadir: Path) -> None:
        self._datadir = datadir.absolute()
        self._process: Optional["Process"] = None
        self._host: str = ""
        self._port: int = -1
        self.started: bool = False

    @property
    def datadir(self) -> Path:
        return self._datadir

    @property
    def host(self) -> str:
        if not self.started:
            self.start()

        assert (
            self._host
        ), "The action server was not properly started (no host available)"
        return self._host

    @property
    def port(self) -> int:
        if not self.started:
            self.start()

        assert (
            self._port > 0
        ), "The action server was not properly started (no port available)"
        return self._port

    @property
    def process(self) -> "Process":
        assert (
            self._process is not None
        ), "The action server was not properly started (process is None)."
        return self._process

    def start(
        self,
        *,
        timeout: int = 10,
        db_file=":memory:",
        actions_sync=False,
        cwd: Optional[Path | str] = None,
        add_shutdown_api: bool = False,
        min_processes: int = 0,
        max_processes: int = 20,
        reuse_processes: bool = False,
        additional_args: Optional[list[str]] = None,
    ) -> None:
        from robocorp.action_server._robo_utils.process import Process
        from robocorp.action_server._settings import is_frozen

        if self.started:
            raise RuntimeError("The action process was already started.")

        self.started = True
        from concurrent.futures import Future

        if actions_sync:
            assert cwd, "cwd must be passed when synchronizing the actions."

        if is_frozen():
            base_args = [sys.executable]
        else:
            base_args = [
                sys.executable,
                "-m",
                "robocorp.action_server",
            ]
        new_args = base_args + [
            "start",
            "--actions-sync=false" if not actions_sync else "--actions-sync=true",
            "--port=0",
            "--verbose",
            f"--datadir={str(self._datadir)}",
            f"--db-file={db_file}",
        ]

        new_args.append(f"--min-processes={min_processes}")
        new_args.append(f"--max-processes={max_processes}")
        if reuse_processes:
            new_args.append("--reuse-processes")

        if additional_args:
            new_args = new_args + additional_args

        env = {}
        if add_shutdown_api:
            env["RC_ADD_SHUTDOWN_API"] = "1"
        process = self._process = Process(new_args, cwd=cwd, env=env)

        compiled = re.compile(r"Uvicorn running on http://([\w.-]+):(\d+)")
        future: Future[Tuple[str, str]] = Future()

        def collect_port_from_stdout(line):
            # Note: this is called in a thread.
            matches = re.findall(compiled, line)

            if matches:
                host, port = matches[0]
                future.set_result((host, port))

        def on_stdout(line):
            if self.SHOW_OUTPUT:
                sys.stdout.write(f"stdout: {line.rstrip()}\n")

        def on_stderr(line):
            # Note: this is called in a thread.
            sys.stderr.write(f"stderr: {line.rstrip()}\n")

        process.on_stderr.register(on_stderr)
        process.on_stdout.register(on_stdout)

        with process.on_stderr.register(collect_port_from_stdout):
            process.start()
            if timeout > 1:
                initial_time = time.monotonic()
                while True:
                    try:
                        host, port = future.result(1)
                        break
                    except TimeoutError:
                        if is_debugger_active():
                            continue
                        if time.monotonic() - initial_time >= timeout:
                            raise TimeoutError()
                        if not process.is_alive():
                            raise RuntimeError(
                                f"The process already exited with returncode: "
                                f"{process.returncode}"
                            )
            else:
                host, port = future.result(timeout)
        assert host
        self._host = host
        assert int(port) > 0, f"Expected port to be > 0. Found: {port}"
        self._port = int(port)

    def stop(self):
        if self._process is not None:
            self._process.stop()
            self._process = None


class ActionServerClient:
    def __init__(self, action_server_process: ActionServerProcess):
        self.action_server_process = action_server_process

    def build_full_url(self, url):
        host = self.action_server_process.host
        port = self.action_server_process.port
        if url.startswith("/"):
            url = url[1:]
        return f"http://{host}:{port}/{url}"

    def get_str(self, url, params: Optional[dict] = None) -> str:
        import requests

        result = requests.get(self.build_full_url(url), params=(params or {}))
        assert result.status_code == 200
        return result.text

    def get_openapi_json(self):
        return self.get_str("openapi.json")

    def get_json(self, url, params: Optional[dict] = None):
        import json

        contents = self.get_str(url, params=params)
        try:
            return json.loads(contents)
        except Exception:
            raise AssertionError(f"Unable to load: {contents!r}")

    def post_get_str(self, url, data):
        import requests

        result = requests.post(self.build_full_url(url), json=data)
        assert result.status_code == 200
        return result.text

    def post_error(self, url, status_code, data=None):
        import requests

        result = requests.post(self.build_full_url(url), json=data or {})
        assert result.status_code == status_code
        return result

    def get_error(self, url, status_code):
        import requests

        result = requests.get(self.build_full_url(url))
        assert result.status_code == status_code


def robocorp_action_server_run(
    cmdline,
    returncode: Union[Literal["error"], int],
    cwd=None,
    additional_env: Optional[Dict[str, str]] = None,
    timeout=None,
) -> CompletedProcess:
    from robocorp.action_server._settings import is_frozen

    if is_frozen():
        # i.e.: The entry point is our own executable.
        return run_command_line(
            [sys.executable] + cmdline, returncode, cwd, additional_env, timeout
        )
    else:
        return run_python_module(
            "robocorp.action_server", cmdline, returncode, cwd, additional_env, timeout
        )


def run_python_module(
    python_module: str,
    cmdline,
    returncode: Union[Literal["error"], int],
    cwd=None,
    additional_env: Optional[Dict[str, str]] = None,
    timeout=None,
) -> CompletedProcess:
    return run_command_line(
        [sys.executable, "-m", python_module] + cmdline,
        returncode=returncode,
        cwd=cwd,
        additional_env=additional_env,
        timeout=timeout,
    )


def run_command_line(
    cmdline,
    returncode: Union[Literal["error"], int],
    cwd=None,
    additional_env: Optional[Dict[str, str]] = None,
    timeout=None,
) -> CompletedProcess:
    cp = os.environ.copy()
    cp["PYTHONPATH"] = os.pathsep.join([x for x in sys.path if x])
    if additional_env:
        cp.update(additional_env)
    result = subprocess.run(
        cmdline, capture_output=True, text=True, env=cp, cwd=cwd, timeout=timeout
    )

    if returncode == "error" and result.returncode:
        return result

    if result.returncode == returncode:
        return result

    env_str = "\n".join(str(x) for x in sorted(cp.items()))

    raise AssertionError(
        f"""Expected returncode: {returncode}. Found: {result.returncode}.
=== stdout:
{result.stdout}

=== stderr:
{result.stderr}

=== Env:
{env_str}

=== Args:
{cmdline}

"""
    )


def check_new_template(
    tmpdir: Path,
    action_server_process: ActionServerProcess,
    client: ActionServerClient,
    verbose: bool = False,
) -> None:
    from robocorp.action_server._settings import is_frozen

    curdir = os.path.abspath(".")
    try:
        os.chdir(str(tmpdir))
        if verbose:
            print(f"is_frozen(): {is_frozen()}")
            print(f"Creating template project in: {tmpdir}")

        output = robocorp_action_server_run(
            ["new", "--name=my_project"], returncode=0, cwd=str(tmpdir)
        )
        if verbose:
            print("Template creation stdout: ", output.stdout)
            print("Template creation stderr: ", output.stderr)

        my_project_dir = tmpdir / "my_project"
        my_project_dir_conda = my_project_dir / "conda.yaml"

        if verbose:
            print(my_project_dir, "exists", my_project_dir.exists())
            print(my_project_dir_conda, "exists", my_project_dir_conda.exists())

        if not my_project_dir_conda.exists():
            raise RuntimeError(f"Expected {my_project_dir_conda} to exist.")

        # Note: timeout is big because it'll use rcc to bootstrap the env here.
        if verbose:
            print("Starting action-server in template project.")
        action_server_process.start(
            db_file="server.db",
            cwd=str(tmpdir / "my_project"),
            actions_sync=True,
            timeout=300,
        )

        if verbose:
            print("Querying action packages available.")

        action_packages = client.get_json("api/actionPackages")
        assert len(action_packages) == 1
        action_package = next(iter(action_packages))
        actions = action_package["actions"]
        action_names = tuple(action["name"] for action in actions)
        assert "compare_time_zones" in action_names

        if verbose:
            print("Using post to call action.")

        found = client.post_get_str(
            "/api/actions/my-project/compare-time-zones/run",
            {
                "user_timezone": "Europe/Helsinki",
                "compare_to_timezones": "America/New_York, Asia/Kolkata",
            },
        )
        assert "Current time in Europe/Helsinki" in found
        assert "Current time in America/New_York" in found
        assert "Current time in Asia/Kolkata" in found

        if verbose:
            print("Test finished with success.")
    finally:
        os.chdir(curdir)


@contextmanager
def make_tmpdir() -> Iterator[Path]:
    import shutil
    import tempfile

    temp_dir = tempfile.mkdtemp()
    try:
        yield Path(temp_dir)
    finally:
        retry = 4
        for i in range(retry):
            try:
                shutil.rmtree(temp_dir)
                break
            except Exception:
                # It's possible that the process takes a bit longer to cleanup,
                # so, wait a bit and retry.
                time.sleep(0.4)
                if i == retry - 1:
                    raise


@contextmanager
def make_action_server_process(tmpdir: Path) -> Iterator[ActionServerProcess]:
    action_server_datadir = tmpdir / ".robocorp_action_server"
    ret = ActionServerProcess(action_server_datadir)
    try:
        yield ret
    finally:
        ret.stop()


@contextmanager
def make_client(
    action_server_process: ActionServerProcess,
) -> Iterator[ActionServerClient]:
    yield ActionServerClient(action_server_process)


def do_selftest():
    print("Running selftest...")
    retcode = 0

    from robocorp.action_server._download_rcc import get_default_rcc_location

    rcc_location = get_default_rcc_location()
    assert rcc_location.exists(), f"Expected rcc to be available in: {rcc_location}."
    print(f"Permissions for: {rcc_location}: {oct(os.stat(rcc_location).st_mode)[-3:]}")
    if not os.access(str(rcc_location), os.X_OK):
        raise AssertionError("Expected rcc to have the executable bit set.")

    try:
        with make_tmpdir() as tmpdir:
            with make_action_server_process(tmpdir) as action_server_process:
                with make_client(action_server_process) as client:
                    check_new_template(
                        tmpdir, action_server_process, client, verbose=True
                    )
    except Exception:
        print("selftest failed with error!")
        import traceback

        traceback.print_exc()
        retcode = 1

    if retcode != 0:
        print("Action server selftest failed!")
    else:
        print("Action server selftest worked!")
    return retcode



import os
import platform
import sys
from pathlib import Path
from typing import Optional

# No real build, just download RCC at this point.

# Note: referenced here and in robocorp.action_server._download_rcc
RCC_VERSION = "17.12.1"

RCC_URLS = {
    "Windows": f"https://downloads.robocorp.com/rcc/releases/v{RCC_VERSION}/windows64/rcc.exe",
    "Darwin": f"https://downloads.robocorp.com/rcc/releases/v{RCC_VERSION}/macos64/rcc",
    "Linux": f"https://downloads.robocorp.com/rcc/releases/v{RCC_VERSION}/linux64/rcc",
}

CURDIR = Path(__file__).parent.absolute()


def _download_rcc(system: Optional[str] = None, target: Optional[str] = None):
    """
    Downloads RCC in the place where the action server expects it.
    """
    import stat
    import urllib.request

    if target:
        rcc_path = Path(target)
    else:
        if sys.platform == "win32":
            rcc_path = CURDIR / "src" / "robocorp" / "action_server" / "bin" / "rcc.exe"
        else:
            rcc_path = CURDIR / "src" / "robocorp" / "action_server" / "bin" / "rcc"

    rcc_url = RCC_URLS[system or platform.system()]

    print(f"Downloading '{rcc_url}' to '{rcc_path}'")

    # Cloudflare seems to be blocking "User-Agent: Python-urllib/3.9".
    # Use a different one as that must be sorted out.
    response = urllib.request.urlopen(
        urllib.request.Request(rcc_url, headers={"User-Agent": "Mozilla"})
    )

    with open(rcc_path, "wb") as stream:
        stream.write(response.read())

    st = os.stat(rcc_path)
    os.chmod(rcc_path, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

    return rcc_path


def build(*args, **kwarg):
    if os.environ.get("ACTION_SERVER_SKIP_DOWNLOAD_IN_BUILD", "").lower().strip() in (
        "1",
        "true",
    ):
        return

    _download_rcc()


if __name__ == "__main__":
    build()



import typing
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator, Optional, Union

from pydantic.dataclasses import dataclass

from robocorp.action_server._database import DBRules

if typing.TYPE_CHECKING:
    from robocorp.action_server._database import Database


_db_rules = DBRules()


@dataclass
class ActionPackage:  # Table name: action_package
    id: str  # primary key (uuid)
    _db_rules.unique_indexes.add("ActionPackage.id")

    name: str  # The name for the action package
    _db_rules.unique_indexes.add("ActionPackage.name")

    # The directory where the action package is (may be stored relative
    # to the datadir or as an absolute path).
    # When relative starts with `./`, otherwise it's absolute.
    directory: str

    # The sha256 hash of the conda.yaml (based only on the actual content
    # not considering spaces and comments).
    conda_hash: str

    # The environment to be used for launches (as json).
    env_json: str


@dataclass
class Action:  # Table name: action
    id: str  # primary key (uuid)
    _db_rules.unique_indexes.add("Action.id")

    action_package_id: str  # foreign key to the action package
    _db_rules.foreign_keys.add("Action.action_package_id")
    _db_rules.indexes.add("Action.action_package_id")

    name: str  # The action name
    docs: str  # Docs for the action

    # File for the action (relative to the directory in the ActionPackage).
    file: str

    lineno: int  # Line for the action
    input_schema: str  # The json content for the schema input
    output_schema: str  # The json content for the schema output

    enabled: bool = True

    is_consequential: Optional[bool] = None


RUN_ID_COUNTER = "run_id"
ALL_COUNTERS = (RUN_ID_COUNTER,)


@dataclass
class Counter:
    id: str  # primary key (counter name -- i.e.: RUN_ID_COUNTER)
    _db_rules.unique_indexes.add("Counter.id")

    value: int  # current value


@dataclass
class Run:
    id: str  # primary key (uuid)
    _db_rules.unique_indexes.add("Run.id")

    status: int  # 0=not run, 1=running, 2=passed, 3=failed
    _db_rules.indexes.add("Run.status")

    action_id: str  # foreign key to the action
    _db_rules.foreign_keys.add("Run.action_id")
    _db_rules.indexes.add("Run.action_id")

    start_time: str  # The time of the run creation.
    run_time: Optional[
        float
    ]  # The time from the run creation to the run finish (in seconds)
    inputs: str  # The json content with the variables used as an input
    result: Optional[str]  # The json content of the output that the run generated
    error_message: Optional[str]  # If the status=failed, this may have an error message

    # The path (relative to the datadir) of the artifacts generated in the run
    relative_artifacts_dir: str

    # We could've made this the primary key, but in theory when (if)
    # we ever have workspaces then this id would need to be scoped to the
    # workspace while the other one would be global.
    numbered_id: int
    _db_rules.unique_indexes.add("Run.numbered_id")


class RunStatus:
    NOT_RUN = 0
    RUNNING = 1
    PASSED = 2
    FAILED = 3


def get_all_model_classes():
    from robocorp.action_server.migrations import Migration

    return [Migration, ActionPackage, Action, Run, Counter]


def get_model_db_rules() -> DBRules:
    return _db_rules


_global_db: Optional["Database"] = None


@contextmanager
def load_db(db_path: Union[Path, str]) -> Iterator["Database"]:
    """
    Loads the database from the given path and initializes the internal
    models, besides setting this db as the global db for the duration
    of the context manager.
    """
    from robocorp.action_server._database import Database

    global _global_db

    if _global_db is not None:
        raise AssertionError("There is already a global initialized database.")

    db = Database(db_path)
    with db.connect():
        db.initialize(get_all_model_classes())
        _global_db = db
        try:
            yield db
        finally:
            _global_db = None


@contextmanager
def create_db(db_path: Union[Path, str]) -> Iterator["Database"]:
    """
    Creates the database and sets this db as the global db for the duration
    of the context manager.
    """
    from robocorp.action_server.migrations import (
        CURRENT_VERSION,
        MIGRATION_ID_TO_NAME,
        Migration,
    )

    with load_db(db_path) as db:
        with db.transaction():
            db.create_tables(get_model_db_rules())
            current_migration = MIGRATION_ID_TO_NAME[CURRENT_VERSION]
            db.insert(Migration(CURRENT_VERSION, current_migration))
            for counter in ALL_COUNTERS:
                db.insert(Counter(counter, 0))
        yield db


def get_db() -> "Database":
    assert _global_db is not None, "DB not initialized"
    return _global_db


def get_action_package_from_action(action: Action) -> ActionPackage:
    db = get_db()
    return db.first(
        ActionPackage,
        "SELECT * FROM action_package WHERE id = ?",
        [action.action_package_id],
    )



"""
Note: this is easy while we're in a single process!

If we ever change the design to support multiple processes we'd need to have a
way to synchronize state across multiple processes.
"""
import threading
import typing
from contextlib import contextmanager
from dataclasses import asdict, dataclass
from typing import Any, Dict, Literal, Optional

if typing.TYPE_CHECKING:
    from ._database import Database
    from ._models import Run


@dataclass(slots=True)
class RunChangeEvent:
    ev: Literal["added", "changed"]
    run: "Run"
    changes: Optional[dict[str, Any]] = None


class RunsState:
    def __init__(self, db: "Database"):
        from ._models import Run

        # Clients that want to register/unregister must use this semaphore
        # to avoid racing conditions.
        #
        # This is done because the expected scenario is the following:
        # 1. The client acquires this semaphore
        # 2. then notifies about existing
        # 3. then registers so that it knows about new runs in the structure
        #
        # In this case, if the client doesn't hold the semaphore himself we
        # could have a racing condition.
        #
        # We could make it an RLock, but then we'd need a wrapper to do the
        # verification on whether it's acquired, so, we use a Semaphore to
        # use the `_value` to do the needed asserts.
        self.semaphore = threading.Semaphore(1)

        # Use dict keys for uniqueness and ordering.
        self._run_listeners: Dict[Any, int] = {}

        # This is an in-memory view of the runs available to be seen by clients.
        # Older runs are in the db, but we don't provide them for the time being.

        # Note: we limit to the latest 200 Runs for now...
        # could be tweaked in the future.
        run_id_to_run = {}
        # order by is important so that we get the latest ones.
        for run in db.all(Run, offset=0, limit=200, order_by="numbered_id DESC"):
            run_id_to_run[run.id] = run

        self._run_id_to_run = run_id_to_run

    def get_current_run_state(self) -> list["Run"]:
        assert (
            self.semaphore._value == 0
        ), "Clients getting the current run state must acquire the semaphore."
        return list(self._run_id_to_run.values())

    def get_run_from_id(self, run_id) -> "Run":
        assert (
            self.semaphore._value == 0
        ), "Clients getting the current run state must acquire the semaphore."
        return self._run_id_to_run[run_id]

    def register(self, listener):
        assert (
            self.semaphore._value == 0
        ), "Clients registering must acquire the semaphore."
        self._run_listeners[listener] = 1

    def unregister(self, listener):
        assert (
            self.semaphore._value == 0
        ), "Clients unregistering must acquire the semaphore."
        self._run_listeners.pop(listener, None)

    def on_run_inserted(self, run: "Run"):
        # Semaphore is acquired internally in this case.
        from ._models import Run

        run_copy = Run(**asdict(run))
        with self.semaphore:
            self._run_id_to_run[run_copy.id] = run_copy
            for listener in self._run_listeners.keys():
                listener(RunChangeEvent("added", run_copy))

    def on_run_changed(self, run: "Run", changes: Dict[str, Any]):
        # Semaphore is acquired internally in this case.
        from ._models import Run

        run_copy = Run(**asdict(run))

        with self.semaphore:
            self._run_id_to_run[run_copy.id] = run_copy
            for listener in self._run_listeners.keys():
                listener(RunChangeEvent("changed", run_copy, changes))


_runs_state: Optional[RunsState] = None


@contextmanager
def use_runs_state_ctx(db: "Database"):
    global _runs_state
    _runs_state = RunsState(db)
    try:
        yield _runs_state
    finally:
        pass
    _runs_state = None


def get_global_runs_state() -> RunsState:
    assert _runs_state
    return _runs_state



from robocorp.action_server._database import Database


def migrate(db: Database) -> None:
    raise RuntimeError(
        f"""Error: 
It seems that this version of the database ({db.db_path}) is too old.
Please erase it and recreate it from scratch."""
    )



import logging
import traceback
from enum import Enum
from typing import Any, Dict, Union

from fastapi import status
from fastapi.exceptions import HTTPException, RequestValidationError
from fastapi.middleware.cors import CORSMiddleware
from fastapi.openapi.constants import REF_PREFIX
from fastapi.openapi.utils import validation_error_response_definition
from fastapi.requests import Request
from fastapi.responses import JSONResponse
from pydantic import BaseModel, ValidationError

LOGGER = logging.getLogger(__name__)


class ErrorCode(str, Enum):
    UNKNOWN = "unknown"
    INTERNAL_ERROR = "internal-error"
    VALIDATION_ERROR = "validation-error"
    TIMEOUT_ERROR = "timeout-error"
    NOT_AVAILABLE = "not-available"
    INVALID_REQUEST = "invalid-request"
    INVALID_VALUE = "invalid-value"


class RequestError(RuntimeError):
    def __init__(self, error_code: ErrorCode, message: str):
        super().__init__()
        self.error_code = error_code
        self.message = message


class ErrorResponse(BaseModel):
    error_code: ErrorCode
    message: str


def _to_response(
    error_code: ErrorCode,
    message: str,
    path: str,
    trace=False,
) -> Dict[str, Any]:
    LOGGER.error("Invalid request: %s [%s]", message, path)
    if trace:
        LOGGER.error(traceback.format_exc())

    return {"error_code": error_code, "message": message}


async def request_error_handler(request: Request, exc: RequestError):
    return JSONResponse(
        status_code=400,
        content=_to_response(
            error_code=exc.error_code,
            message=exc.message,
            path=str(request.url.path),
        ),
    )


async def http_error_handler(
    request: Request,
    exc: HTTPException,
) -> JSONResponse:
    return JSONResponse(
        status_code=exc.status_code,
        content=_to_response(
            error_code=ErrorCode.UNKNOWN,
            message=str(exc.detail),
            path=str(request.url.path),
        ),
    )


async def http422_error_handler(
    request: Request,
    exc: Union[RequestValidationError, ValidationError],
) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
        content=_to_response(
            error_code=ErrorCode.VALIDATION_ERROR,
            message="\n".join(str(err) for err in exc.errors()),
            path=str(request.url.path),
        ),
    )


async def http500_error_handler(
    request: Request,
    exc: Exception,
) -> JSONResponse:
    response = JSONResponse(
        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        content=_to_response(
            error_code=ErrorCode.INTERNAL_ERROR,
            message=str(exc),
            path=str(request.url.path),
            trace=True,
        ),
    )

    # CORSMiddleware is not used for unhandled server exceptions
    # in FastAPI/Starlette, so we set it manually here
    origin = request.headers.get("origin")
    if origin:
        cors = CORSMiddleware(
            app=request.app,
            allow_origins=["*"],
            allow_methods=["*"],
            allow_headers=["*"],
            allow_credentials=True,
        )
        response.headers.update(cors.simple_headers)
        if cors.allow_all_origins and "cookie" in request.headers:
            response.headers["Access-Control-Allow-Origin"] = origin
        elif not cors.allow_all_origins and cors.is_allowed_origin(origin=origin):
            response.headers["Access-Control-Allow-Origin"] = origin
            response.headers.add_vary_header("Origin")

    return response


validation_error_response_definition["properties"] = {
    "errors": {
        "title": "Errors",
        "type": "array",
        "items": {"$ref": f"{REF_PREFIX}ValidationError"},
    },
}



import itertools
from typing import Literal, Optional
from uuid import uuid4

_counters: dict = {}


def get_counter(counter_name: Optional[str] = None):
    try:
        return _counters[counter_name]
    except KeyError:
        _counters[counter_name] = itertools.count()
        return _counters[counter_name]


def gen_uuid(counter_name: Optional[Literal["action_package", "action", "run"]] = None):
    if not counter_name:
        prefix = ""
    elif counter_name == "action_package":
        prefix = "ap-"
    else:
        prefix = f"{str(counter_name)[0:3]}-"
    return f"{prefix}{next(get_counter(counter_name)):03d}-{uuid4()}"



from logging import getLogger

logger = getLogger(__name__)


class OnExitContextManager:
    def __init__(self, on_exit):
        self.on_exit = on_exit

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.on_exit()


class Callback:
    """
    Note that it's thread safe to register/unregister callbacks while callbacks
    are being notified, but it's not thread-safe to register/unregister at the
    same time in multiple threads.
    """

    def __init__(self, reversed=False, raise_exceptions=False):
        self.raise_exceptions = raise_exceptions
        self._reversed = reversed
        self._callbacks = ()

    def register(self, callback):
        self._callbacks = self._callbacks + (callback,)

        # Enable using as a context manager to automatically call the unregister.
        return OnExitContextManager(lambda: self.unregister(callback))

    def unregister(self, callback):
        self._callbacks = tuple(x for x in self._callbacks if x != callback)

    def __len__(self):
        return len(self._callbacks)

    def __call__(self, *args, **kwargs):
        if self._reversed:
            iter_in = reversed(self._callbacks)
        else:
            iter_in = self._callbacks
        for c in iter_in:
            try:
                c(*args, **kwargs)
            except Exception:
                logger.exception(f"Error calling: {c}.")
                if self.raise_exceptions:
                    raise



import json
import logging
import os
import subprocess
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from subprocess import CalledProcessError, TimeoutExpired
from typing import Dict, Iterator, List, Optional, Tuple

from robocorp.action_server._protocols import ActionResult, RCCActionResult, Sentinel
from robocorp.action_server._robo_utils.constants import NULL

log = logging.getLogger(__name__)


def create_hash(contents: str) -> str:
    import hashlib

    sha256_hash = hashlib.sha256()
    sha256_hash.update(contents.encode("utf-8"))
    return sha256_hash.hexdigest()


RCC_CLOUD_ROBOT_MUTEX_NAME = "rcc_cloud_activity"
RCC_CREDENTIALS_MUTEX_NAME = "rcc_credentials"


class EnvInfo(object):
    def __init__(self, env: Dict[str, str]):
        self.env = env


def as_str(s) -> str:
    if isinstance(s, bytes):
        return s.decode("utf-8", "replace")
    return str(s)


class Rcc(object):
    def __init__(self, rcc_location: Path, robocorp_home: Path):
        self._rcc_location = rcc_location
        self._robocorp_home = robocorp_home
        self.config_location = os.environ.get(
            "RC_ACTION_SERVER_RCC_CONFIG_LOCATION", ""
        )

    def _compute_env(self):
        env = os.environ.copy()
        env.pop("PYTHONPATH", "")
        env.pop("PYTHONHOME", "")
        env.pop("VIRTUAL_ENV", "")
        env["PYTHONIOENCODING"] = "utf-8"
        env["PYTHONUNBUFFERED"] = "1"

        env["ROBOCORP_HOME"] = str(self._robocorp_home)
        return env

    def _compute_launch_args_and_kwargs(
        self, cwd, env, args, stderr=Sentinel.SENTINEL
    ) -> Tuple[list, dict]:
        from robocorp.action_server._robo_utils.process import build_subprocess_kwargs

        if stderr is Sentinel.SENTINEL:
            stderr = subprocess.PIPE

        kwargs: dict = build_subprocess_kwargs(cwd, env, stderr=stderr)
        rcc_location = str(self._rcc_location)
        args = [rcc_location] + args + ["--controller", "ActionServer"]
        return args, kwargs

    def _run_rcc(
        self,
        args: List[str],
        timeout: float = 35,
        error_msg: str = "",
        mutex_name=None,
        cwd: Optional[str] = None,
        log_errors=True,
        stderr=Sentinel.SENTINEL,
        show_interactive_output: bool = False,
        hide_in_log: Optional[str] = None,
    ) -> RCCActionResult:
        """
        Returns an ActionResult where the result is the stdout of the executed command.

        :param log_errors:
            If false, errors won't be logged (i.e.: should be false when errors
            are expected).

        :param stderr:
            If given sets the stderr redirection (by default it's subprocess.PIPE,
            but users could change it to something as subprocess.STDOUT).
        """
        from subprocess import check_output, list2cmdline

        from robocorp.action_server._robo_utils.process import check_output_interactive

        env = self._compute_env()
        robocorp_home = env["ROBOCORP_HOME"]

        args, kwargs = self._compute_launch_args_and_kwargs(cwd, env, args, stderr)
        cmdline = list2cmdline([str(x) for x in args])

        try:
            if mutex_name:
                from ._robo_utils.system_mutex import timed_acquire_mutex
            else:
                timed_acquire_mutex = NULL
            with timed_acquire_mutex(mutex_name, timeout=15):
                if logging.root.level <= logging.DEBUG:
                    msg = f"Running: {cmdline}"
                    if hide_in_log:
                        msg = msg.replace(hide_in_log, "<HIDDEN_IN_LOG>")

                    log.debug(msg)

                curtime = time.monotonic()

                boutput: bytes
                # We have 2 main modes here: one in which we can print the output
                # interactively while the command is running and another where
                # we only print if some error happened.
                if not show_interactive_output:
                    boutput = check_output(args, timeout=timeout, **kwargs)
                else:

                    def on_output(content):
                        try:
                            sys.stderr.buffer.write(content)
                        except BaseException:
                            log.exception("Error reporting interactive output.")

                    boutput = check_output_interactive(
                        args,
                        timeout=timeout,
                        on_stderr=on_output,
                        on_stdout=on_output,
                        **kwargs,
                    )

        except CalledProcessError as e:
            stdout = as_str(e.stdout)
            stderr = as_str(e.stderr)
            msg = (
                f"Error running: {cmdline}.\nROBOCORP_HOME: {robocorp_home}\n\n"
                f"Stdout: {stdout}\nStderr: {stderr}"
            )
            if hide_in_log:
                msg = msg.replace(hide_in_log, "<HIDDEN_IN_LOG>")

            if log_errors:
                log.exception(msg)
            if not error_msg:
                return RCCActionResult(cmdline, success=False, message=msg)
            else:
                additional_info = [error_msg]
                if stdout or stderr:
                    if stdout and stderr:
                        additional_info.append("\nDetails: ")
                        additional_info.append("\nStdout")
                        additional_info.append(stdout)
                        additional_info.append("\nStderr")
                        additional_info.append(stderr)

                    elif stdout:
                        additional_info.append("\nDetails: ")
                        additional_info.append(stdout)

                    elif stderr:
                        additional_info.append("\nDetails: ")
                        additional_info.append(stderr)

                return RCCActionResult(
                    cmdline, success=False, message="".join(additional_info)
                )

        except TimeoutExpired:
            msg = f"Timed out ({timeout}s elapsed) when running: {cmdline}"
            log.exception(msg)
            return RCCActionResult(cmdline, success=False, message=msg)

        except Exception:
            msg = f"Error running: {cmdline}"
            log.exception(msg)
            return RCCActionResult(cmdline, success=False, message=msg)

        output = boutput.decode("utf-8", "replace")

        do_log_as_info = (
            log_errors and logging.root.level < logging.INFO
        ) or logging.root.level <= logging.DEBUG

        if do_log_as_info:
            elapsed = time.monotonic() - curtime
            msg = f"Output from: {cmdline} (took: {elapsed:.2f}s): {output}"
            if hide_in_log:
                msg = msg.replace(hide_in_log, "<HIDDEN_IN_LOG>")
            log.info(msg)

        return RCCActionResult(cmdline, success=True, message=None, result=output)

    def create_env_and_get_vars(
        self, conda_yaml: Path, conda_hash: str
    ) -> ActionResult[EnvInfo]:
        args = [
            "holotree",
            "variables",
            "--space",
            conda_hash,
            str(conda_yaml),
        ]
        args.append("--json")
        timeout = 60 * 60  # Wait up to 1 hour for the env...
        ret = self._run_rcc(
            args,
            mutex_name=RCC_CLOUD_ROBOT_MUTEX_NAME,
            cwd=str(conda_yaml.parent),
            timeout=timeout,  # Creating the env may be really slow!
            show_interactive_output=logging.root.level <= logging.DEBUG,
        )

        def return_failure(msg: Optional[str]) -> ActionResult[EnvInfo]:
            log.critical(
                (
                    "Unable to create environment from:\n%s\n"
                    "To recreate the environment, please change the related conda yaml"
                    "\nor restart VSCode to retry with the same conda yaml contents."
                ),
                conda_yaml,
            )

            if not msg:
                msg = "<unknown reason>"
            log.critical(msg)
            action_result: ActionResult[EnvInfo] = ActionResult(False, msg, None)
            return action_result

        if not ret.success:
            return return_failure(ret.message)

        contents: Optional[str] = ret.result
        if not contents:
            return return_failure("Unable to get output when getting environment.")

        environ = {}
        for entry in json.loads(contents):
            key = str(entry["key"])
            value = str(entry["value"])
            if key:
                environ[key] = value

        if "CONDA_PREFIX" not in environ:
            msg = f"Did not find CONDA_PREFIX in {environ}"
            return return_failure(msg)

        if "PYTHON_EXE" not in environ:
            msg = f"Did not find PYTHON_EXE in {environ}"
            return return_failure(msg)

        return ActionResult(True, None, EnvInfo(environ))

    def pull(self, url: str, directory: str) -> ActionResult[str]:
        args = ["pull", url, "--directory", directory]
        ret = self._run_rcc(args, mutex_name=RCC_CLOUD_ROBOT_MUTEX_NAME)
        if not ret.success:
            return ActionResult(False, ret.message, None)
        return ActionResult(True, None, ret.result)

    def _add_config_to_args(self, args: List[str]) -> List[str]:
        config_location = self.config_location
        if config_location:
            args.append("--config")
            args.append(config_location)
        return args

    def feedack_metric(self, name, value="+1") -> None:
        env = self._compute_env()

        args = ["feedback", "metric", "-t", "action-server", "-n", name, "-v", value]
        self._add_config_to_args(args)
        cwd = None
        args, kwargs = self._compute_launch_args_and_kwargs(cwd, env, args)
        try:
            subprocess.Popen(args, **kwargs)
        except BaseException:
            log.exception("Error submitting feedback.")


_rcc: Optional["Rcc"] = None


@contextmanager
def initialize_rcc(rcc_location: Path, robocorp_home: Path) -> Iterator[Rcc]:
    global _rcc

    rcc = Rcc(rcc_location, robocorp_home)
    _rcc = rcc
    try:
        yield rcc
    finally:
        _rcc = None


def get_rcc() -> Rcc:
    assert _rcc is not None, "RCC not initialized"
    return _rcc



from enum import Enum
from typing import Any, Generic, Optional, TypedDict, TypeVar

T = TypeVar("T")


class Sentinel(Enum):
    SENTINEL = 0


class ActionResultDict(TypedDict):
    success: bool
    message: Optional[
        str
    ]  # if success == False, this can be some message to show to the user
    result: Any


class ActionResult(Generic[T]):
    success: bool
    message: Optional[
        str
    ]  # if success == False, this can be some message to show to the user
    result: Optional[T]

    def __init__(
        self, success: bool, message: Optional[str] = None, result: Optional[T] = None
    ):
        self.success = success
        self.message = message
        self.result = result

    def as_dict(self) -> ActionResultDict:
        return {"success": self.success, "message": self.message, "result": self.result}

    def __str__(self):
        return (
            f"ActionResult(success={self.success!r}, message={self.message!r}, "
            f"result={self.result!r})"
        )

    __repr__ = __str__


class RCCActionResult(ActionResult[str]):
    # A string-representation of the command line.
    command_line: str

    def __init__(
        self,
        command_line: str,
        success: bool,
        message: Optional[str] = None,
        result: Optional[str] = None,
    ):
        ActionResult.__init__(self, success, message, result)
        self.command_line = command_line


def check_implements(x: T) -> T:
    """
    Helper to check if a class implements some protocol.

    :important: It must be the last method in a class due to
                https://github.com/python/mypy/issues/9266

        Example:

    def __typecheckself__(self) -> None:
        _: IExpectedProtocol = check_implements(self)

    Mypy should complain if `self` is not implementing the IExpectedProtocol.
    """
    return x



import argparse
import logging
import os.path
import sys
from pathlib import Path
from typing import Optional, Union

from . import __version__

log = logging.getLogger(__name__)


# def _write_schema(path: Optional[str]):
#     from fastapi.openapi.utils import get_openapi
#     from ._app import get_app
#     app = get_app()
#     schema = get_openapi(
#         title=app.title,
#         version=app.version,
#         openapi_version=app.openapi_version,
#         description=app.description,
#         routes=app.routes,
#     )
#
#     if path is None:
#         print(json.dumps(schema, indent=4))
#     else:
#         output = Path(path)
#         output.parent.mkdir(parents=True, exist_ok=True)
#         with open(output, "w", encoding="utf-8") as file:
#             json.dump(schema, file, indent=4)


def _add_data_args(parser, defaults):
    parser.add_argument(
        "-d",
        "--datadir",
        metavar="PATH",
        default="",
        help=(
            "Directory to store the data for operating the actions server "
            "(by default a datadir will be generated based on the current directory)."
        ),
    )
    parser.add_argument(
        "--db-file",
        help=(
            "The name of the database file, relative to the datadir "
            "(default: %(default)s)"
        ),
        default=defaults["db_file"],
    )


def _add_verbose_args(parser, defaults):
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Be more talkative (default: %(default)s)",
    )


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def _create_parser():
    from ._settings import Settings

    defaults = Settings.defaults()
    base_parser = argparse.ArgumentParser(
        prog="action-server",
        description="Robocorp Action Server",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    subparsers = base_parser.add_subparsers(dest="command")

    # Starts the server
    start_parser = subparsers.add_parser(
        "start",
        help=(
            "Starts the Robocorp Action Server (importing the actions in the "
            "current directory by default)."
        ),
    )

    start_parser.add_argument(
        "-a",
        "--address",
        default=defaults["address"],
        help="Server address (default: %(default)s)",
    )
    start_parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=defaults["port"],
        help="Server port (default: %(default)s)",
    )
    start_parser.add_argument(
        "--expose",
        action="store_true",
        help="Expose the server to the world",
    )
    start_parser.add_argument(
        "--server-url",
        help=(
            "Explicit server url to be defined in the OpenAPI spec 'servers' section."
            "Defaults to the localhost url."
        ),
        default=None,
    )
    start_parser.add_argument(
        "--expose-allow-reuse",
        dest="expose_allow_reuse",
        action="store_true",
        help="Always answer yes to expose reuse confirmation",
    )
    start_parser.add_argument(
        "--api-key",
        dest="api_key",
        help=(
            'Adds authentication. Pass it as `{"Authorization": "Bearer <API_KEY>"}` '
            "header. Pass `--api-key None` to disable authentication."
        ),
        default=None,
    )
    start_parser.add_argument(
        "--actions-sync",
        type=str2bool,
        help=(
            "By default the actions will be synchronized (added/removed) given the "
            "directories provided (if not specified the current directory is used). To "
            "start without synchronizing it's possible to use `--actions-sync=false`"
        ),
        default=True,
    )
    start_parser.add_argument(
        "--dir",
        metavar="PATH",
        help="By default, when starting, actions will be collected from the current "
        "directory to serve, but it's also possible to use `--dir` to load actions "
        "from a different directory",
        action="append",
    )

    start_parser.add_argument(
        "--min-processes",
        type=int,
        help=(
            "The minimum number of action processes that should always be kept alive, "
            "ready to process any incoming request."
        ),
        default=1,
    )
    start_parser.add_argument(
        "--max-processes",
        type=int,
        help=(
            "The maximum number of processes that may be created to handle the actions."
        ),
        default=20,
    )
    start_parser.add_argument(
        "--reuse-processes",
        action="store_true",
        help=(
            "By default actions are run once and then after the action runs the "
            "process that ran the action exits. This can be changed by using "
            "--reuse-processes. With this flag, after running the action instead of "
            "exiting the process will be available to run another action in the same "
            "process (note that in this case care must be taken so that memory leakage "
            "does not happen in the action and that global state from one run does not "
            "interfere with a subsequent run)."
        ),
    )

    _add_data_args(start_parser, defaults)
    _add_verbose_args(start_parser, defaults)

    # Import
    import_parser = subparsers.add_parser(
        "import",
        help="Imports an Action Package and exits",
    )

    import_parser.add_argument(
        "--dir",
        metavar="PATH",
        help="Can be used to import an action package from the local filesystem",
        action="append",
    )

    _add_data_args(import_parser, defaults)
    _add_verbose_args(import_parser, defaults)

    # Download RCC
    rcc_parser = subparsers.add_parser(
        "download-rcc",
        help=(
            "Downloads RCC (by default to the location required by the "
            "Robocorp Action Server)"
        ),
    )

    rcc_parser.add_argument(
        "--file",
        metavar="PATH",
        help="Target file to where RCC should be downloaded",
        nargs="?",
    )

    # New project from template
    new_parser = subparsers.add_parser(
        "new",
        help="Bootstrap new project from template",
    )
    new_parser.add_argument(
        "--name",
        help="Name for the project",
    )
    _add_data_args(new_parser, defaults)
    _add_verbose_args(new_parser, defaults)

    # Schema
    # schema_parser = subparsers.add_parser(
    #     "schema",
    #     help="Prints the schema and exits",
    # )
    #
    # schema_parser.add_argument(
    #     "--file",
    #     metavar="PATH",
    #     help=(
    #         "Write openapi.json schema and exit (if path is given the schema "
    #         "is written to that file instead of stdout)"
    #     ),
    #     nargs="?",
    # )

    # Version
    subparsers.add_parser(
        "version",
        help="Prints the version and exits",
    )

    # Migration
    migration_parser = subparsers.add_parser(
        "migrate",
        help="Makes a database migration (if needed) and exits",
    )
    _add_data_args(migration_parser, defaults)
    _add_verbose_args(migration_parser, defaults)

    return base_parser


def main(args: Optional[list[str]] = None, *, exit=True) -> int:
    if args is None:
        args = sys.argv[1:]

    if not args:
        if os.environ.get(
            "RC_ACTION_SERVER_FORCE_DOWNLOAD_RCC", ""
        ).strip().lower() in (
            "1",
            "true",
        ):
            print(
                "As RC_ACTION_SERVER_FORCE_DOWNLOAD_RCC is set and no arguments were "
                "passed, rcc will be downloaded."
            )

            from robocorp.action_server._download_rcc import download_rcc

            download_rcc(force=True)

        if os.environ.get("RC_ACTION_SERVER_DO_SELFTEST", "").strip().lower() in (
            "1",
            "true",
        ):
            from . import _selftest

            print(
                "As RC_ACTION_SERVER_DO_SELFTEST is set and no arguments were passed, "
                "a selftest will be run."
            )

            sys.exit(_selftest.do_selftest())

    retcode = _main_retcode(args, exit=exit)
    if exit:
        sys.exit(retcode)
    return retcode


def _setup_stdout_logging(log_level):
    from logging import StreamHandler

    stream_handler = StreamHandler()
    stream_handler.setLevel(log_level)
    if log_level == logging.DEBUG:
        formatter = logging.Formatter(
            "%(asctime)s [%(levelname)s] %(message)s", datefmt="[%Y-%m-%d %H:%M:%S]"
        )
    else:
        formatter = logging.Formatter("%(message)s", datefmt="[%X]")

    stream_handler.setFormatter(formatter)
    logger = logging.root
    logger.addHandler(stream_handler)


def _setup_logging(datadir: Path, log_level):
    from logging.handlers import RotatingFileHandler

    log_file = str(datadir / "server_log.txt")
    log.info(f"Logs may be found at: {log_file}.")
    rotating_handler = RotatingFileHandler(
        log_file, maxBytes=1_000_000, backupCount=3, encoding="utf-8"
    )
    rotating_handler.setLevel(log_level)
    rotating_handler.setFormatter(
        logging.Formatter(
            "%(asctime)s [%(levelname)s] %(message)s", datefmt="[%Y-%m-%d %H:%M:%S]"
        )
    )
    logger = logging.root
    logger.addHandler(rotating_handler)


def _main_retcode(args: Optional[list[str]], exit) -> int:
    from robocorp.action_server._settings import is_frozen

    from ._download_rcc import download_rcc
    from ._rcc import initialize_rcc
    from ._robo_utils.auth import generate_api_key
    from ._robo_utils.system_mutex import SystemMutex
    from ._runs_state_cache import use_runs_state_ctx

    if args is None:
        args = sys.argv[1:]

    if args and args[0] == "server-expose":
        # The process is being called by to make the server expose.
        # Internal usage only, so, don't even do argument parsing
        # for it.
        from . import _server_expose

        _server_expose.main(*args[1:])
        return 0

    parser = _create_parser()
    base_args = parser.parse_args(args)

    command = base_args.command
    if not command:
        parser.print_help()
        return 0

    if command == "version":
        print(__version__)
        return 0

    # if command == "schema":
    # This doesn't work at this point because we have to register the
    # actions first for it to work.
    #     file = base_args.file
    #     _write_schema(file)
    #     return

    if command == "download-rcc":
        download_rcc(target=base_args.file, force=True)
        return 0

    if command not in (
        "migrate",
        "import",
        "start",
        "new",
    ):
        print(f"Unexpected command: {command}.", file=sys.stderr)
        return 1

    # Log to stdout.
    log_level = logging.DEBUG if base_args.verbose else logging.INFO

    logger = logging.root
    logger.setLevel(log_level)

    _setup_stdout_logging(log_level)

    from ._settings import setup_settings

    with setup_settings(base_args) as settings:
        settings.datadir.mkdir(parents=True, exist_ok=True)
        robocorp_home = settings.datadir / ".robocorp_home"
        robocorp_home.mkdir(parents=True, exist_ok=True)

        with initialize_rcc(download_rcc(force=False), robocorp_home) as rcc:
            if command == "new":
                from ._new_project import create_new_project

                create_new_project(directory=base_args.name)
                return 0

            mutex = SystemMutex("action_server.lock", base_dir=str(settings.datadir))
            if not mutex.get_mutex_aquired():
                print(
                    f"An action server is already started in this datadir ({settings.datadir})."
                    f"\nPlease exit it before starting a new one."
                    f"\nInformation on mutex holder:\n"
                    f"{mutex.mutex_creation_info}",
                    file=sys.stderr,
                )
                return 1

            # Log to file in datadir, always in debug mode
            # (only after lock is in place as multiple loggers to the same
            # file would be troublesome).
            _setup_logging(settings.datadir, log_level)

            try:
                db_path: Union[Path, str]
                if settings.db_file != ":memory:":
                    db_path = settings.datadir / settings.db_file
                else:
                    db_path = settings.db_file

                from robocorp.action_server._models import create_db, load_db
                from robocorp.action_server.migrations import (
                    db_migration_pending,
                    migrate_db,
                )

                is_new = db_path == ":memory:" or not os.path.exists(db_path)

                if is_new:
                    log.info(
                        "Database file does not exist. Creating it at: %s", db_path
                    )
                    use_db_ctx = create_db
                else:
                    use_db_ctx = load_db

                if command == "migrate":
                    if db_path == ":memory:":
                        print(
                            "Cannot do migration of in-memory databases",
                            file=sys.stderr,
                        )
                        return 1
                    if not migrate_db(db_path):
                        return 1
                    return 0
                else:
                    if is_frozen():
                        cmdline = "action-server"
                    else:
                        cmdline = "python -m robocorp.action_server"

                    if not is_new and db_migration_pending(db_path):
                        print(
                            f"""It was not possible to start the server because a 
database migration is required to use with this version of the
Robocorp Action Server.

Please run the command:

{cmdline} migrate

To migrate the database to the current version
-- or start from scratch by erasing the file: 
{db_path}
"""
                        )
                        return 1

                with use_db_ctx(db_path) as db:
                    from . import _actions_import

                    if command == "import":
                        if not base_args.dir:
                            base_args.dir = ["."]

                        for action_package_dir in base_args.dir:
                            _actions_import.import_action_package(
                                settings.datadir, os.path.abspath(action_package_dir)
                            )
                        return 0

                    elif command == "start":
                        # start imports the current directory by default
                        # (unless --actions-sync=false is specified).
                        log.info("Synchronize actions: %s", base_args.actions_sync)

                        rcc.feedack_metric("action-server.started", __version__)

                        if base_args.actions_sync:
                            if not base_args.dir:
                                base_args.dir = ["."]

                            for action_package_dir in base_args.dir:
                                _actions_import.import_action_package(
                                    settings.datadir,
                                    os.path.abspath(action_package_dir),
                                    disable_not_imported=base_args.actions_sync,
                                )

                        with use_runs_state_ctx(db):
                            from ._server import start_server

                            settings.artifacts_dir.mkdir(parents=True, exist_ok=True)

                            expose_session = None
                            if base_args.expose:
                                from ._server_expose import read_expose_session_json

                                expose_session = read_expose_session_json(
                                    datadir=str(settings.datadir)
                                )
                                if expose_session and not base_args.expose_allow_reuse:
                                    confirm = input(
                                        f"Resume previous expose URL {expose_session.url} Y/N? [Y] "
                                    )
                                    if confirm.lower() == "y" or confirm == "":
                                        log.debug("Resuming previous expose session")
                                    else:
                                        expose_session = None

                            api_key = base_args.api_key
                            if api_key is None:
                                # reuse the previously exposed api key
                                if expose_session and expose_session.api_key:
                                    api_key = expose_session.api_key
                                else:
                                    api_key = generate_api_key()

                            start_server(
                                expose=base_args.expose,
                                api_key=api_key,
                                expose_session=expose_session.expose_session
                                if expose_session
                                else None,
                            )
                            return 0

                    else:
                        print(f"Unexpected command: {command}.", file=sys.stderr)
                        return 1
            finally:
                mutex.release_mutex()


if __name__ == "__main__":
    main()



import logging
from contextlib import contextmanager


class CustomHandler(logging.Handler):
    def __init__(self, level=logging.DEBUG):
        from . import callback

        super().__init__(level)
        self.on_record = callback.Callback()

    def emit(self, record):
        # Define the behavior to handle log records here
        self.on_record(record)


@contextmanager
def add_handler_callback(logger, on_log_emit, level=None):
    """
    Utility to add a handler in a context and remove it afterwards.
    """
    custom_handler = CustomHandler(level if level is not None else logging.DEBUG)
    with custom_handler.on_record.register(on_log_emit):
        initial_level = None
        if level is not None:
            initial_level = logger.level
            logger.setLevel(level)
        logger.addHandler(custom_handler)
        try:
            yield
        finally:
            logger.removeHandler(custom_handler)
            if initial_level is not None:
                logger.setLevel(initial_level)



import asyncio
import logging
import os
import socket
import subprocess
import sys
from functools import partial
from typing import Dict, Optional

from ._server_expose import get_expose_session_payload, read_expose_session_json

log = logging.getLogger(__name__)


def _name_as_summary(name):
    return name.replace("_", " ").title()


def _name_to_url(name):
    return name.replace("_", "-")


def start_server(
    expose: bool, api_key: str | None = None, expose_session: str | None = None
) -> None:
    from dataclasses import asdict

    import docstring_parser
    import uvicorn
    from fastapi.staticfiles import StaticFiles
    from starlette.requests import Request
    from starlette.responses import HTMLResponse

    from . import _actions_process_pool, _actions_run
    from ._api_action_package import action_package_api_router
    from ._api_run import run_api_router
    from ._app import get_app
    from ._models import Action, ActionPackage, get_db
    from ._server_websockets import websocket_api_router
    from ._settings import get_settings

    settings = get_settings()

    settings_dict = asdict(settings)
    settings_str = "\n".join(f"    {k} = {v!r}" for k, v in settings_dict.items())
    log.debug(f"Starting server. Settings:\n{settings_str}")

    app = get_app()

    artifacts_dir = settings.artifacts_dir

    app.mount(
        "/artifacts",
        StaticFiles(directory=artifacts_dir),
        name="artifacts",
    )

    db = get_db()
    action: Action
    action_package_id_to_action_package: Dict[str, ActionPackage] = dict(
        (action_package.id, action_package) for action_package in db.all(ActionPackage)
    )

    actions = db.all(Action)
    for action in actions:
        if not action.enabled:
            # Disabled actions should not be registered.
            continue

        doc_desc: Optional[str] = ""
        if action.docs:
            try:
                parsed = docstring_parser.parse(action.docs)
                doc_desc = parsed.long_description or parsed.short_description
            except Exception:
                log.exception("Error parsing docstring: %s", action.docs)
                doc_desc = str(action.docs or "")

        if not doc_desc:
            doc_desc = ""

        action_package = action_package_id_to_action_package.get(
            action.action_package_id
        )
        if not action_package:
            log.critical("Unable to find action package: %s", action.action_package_id)
            continue

        app.add_api_route(
            f"/api/actions/{_name_to_url(action_package.name)}/{_name_to_url(action.name)}/run",
            _actions_run.generate_func_from_action(action),
            name=action.name,
            summary=_name_as_summary(action.name),
            description=doc_desc,
            operation_id=action.name,
            methods=["POST"],
            openapi_extra={
                "x-openai-isConsequential": action.is_consequential,
            }
            if action.is_consequential is not None
            else None,
        )

    if os.getenv("RC_ADD_SHUTDOWN_API", "").lower() in ("1", "true"):

        def shutdown(timeout: int = -1):
            import _thread
            import time

            from robocorp.action_server._robo_utils.run_in_thread import run_in_thread

            _thread.interrupt_main()

            def interrupt_again():
                time.sleep(timeout)
                _thread.interrupt_main()

            if timeout > 0:
                run_in_thread(interrupt_again, daemon=True)

        app.add_api_route("/api/shutdown/", shutdown, methods=["POST"])

    app.include_router(run_api_router)
    app.include_router(action_package_api_router)
    app.include_router(websocket_api_router)

    @app.get("/config", include_in_schema=False)
    async def serve_config():
        payload = {"expose_url": False}

        if expose:
            current_expose_session = read_expose_session_json(
                datadir=str(settings.datadir)
            )

            expose_session_payload = (
                get_expose_session_payload(current_expose_session.expose_session)
                if current_expose_session
                else None
            )

            if expose_session_payload:
                payload[
                    "expose_url"
                ] = f"https://{expose_session_payload.sessionId}.{settings.expose_url}"

        return payload

    @app.get("/base_log.html", response_class=HTMLResponse)
    async def serve_log_html(request: Request):
        from robocorp.log import _index_v3 as index

        return HTMLResponse(index.FILE_CONTENTS["index.html"])

    async def serve_index(request: Request):
        from ._static_contents import FILE_CONTENTS

        return HTMLResponse(FILE_CONTENTS["index.html"])

    index_routes = ["/", "/runs/{full_path:path}", "/actions/{full_path:path}"]
    for index_route in index_routes:
        app.add_api_route(
            index_route,
            serve_index,
            response_class=HTMLResponse,
            include_in_schema=False,
        )

    expose_subprocess = None

    def expose_later(loop):
        from robocorp.action_server._settings import is_frozen

        nonlocal expose_subprocess

        if not server.started:
            loop.call_later(1 / 15.0, partial(expose_later, loop))
            return

        port = settings.port if settings.port != 0 else None
        host = settings.address
        if port is None:
            sockets_ipv4 = [
                s for s in server.servers[0].sockets if s.family == socket.AF_INET
            ]
            if len(sockets_ipv4) == 0:
                raise Exception("Unable to find a port to expose")
            sockname = sockets_ipv4[0].getsockname()
            host = sockname[0]
            port = sockname[1]

        parent_pid = os.getpid()

        if is_frozen():
            # The executable is 'action-server.exe'.
            args = [sys.executable]
        else:
            # The executable is 'python'.
            args = [
                sys.executable,
                "-m",
                "robocorp.action_server",
            ]

        args += [
            "server-expose",
            str(parent_pid),
            str(port),
            "" if not settings.verbose else "v",
            host,
            settings.expose_url,
            settings.datadir,
            str(api_key),
            str(expose_session),
        ]
        expose_subprocess = subprocess.Popen(args)

    async def _on_startup():
        if expose:
            loop = asyncio.get_event_loop()
            loop.call_later(1 / 15.0, partial(expose_later, loop))

    def _on_shutdown():
        import psutil

        log.info("Stopping action server...")
        from robocorp.action_server._robo_utils.process import (
            kill_process_and_subprocesses,
        )

        expose_pid = None
        if expose_subprocess is not None:
            log.info("Shutting down expose subprocess: %s", expose_subprocess.pid)
            expose_pid = expose_subprocess.pid
            kill_process_and_subprocesses(expose_pid)

        p = psutil.Process(os.getpid())
        try:
            children_processes = list(p.children(recursive=True))
        except Exception:
            log.exception("Error listing subprocesses.")

        for child in children_processes:
            if child.pid != expose_pid:  # If it's still around, don't kill it again.
                log.info("Killing sub-process when exiting action server: %s", child)
                try:
                    kill_process_and_subprocesses(child.pid)
                except Exception:
                    log.exception("Error killing subprocess: %s", child.pid)

    app.add_event_handler("startup", _on_startup)
    app.add_event_handler("shutdown", _on_shutdown)

    with _actions_process_pool.setup_actions_process_pool(
        settings, action_package_id_to_action_package, actions
    ):
        kwargs = settings.to_uvicorn()
        config = uvicorn.Config(app=app, **kwargs)
        server = uvicorn.Server(config)
        asyncio.run(server.serve())



import datetime
import inspect
import json
import logging
import time
import typing
from typing import Annotated, Any, Dict, Optional

from fastapi.params import Header, Param
from pydantic import BaseModel

if typing.TYPE_CHECKING:
    from ._models import Action, Run

log = logging.getLogger(__name__)


_spec_api_type_to_python_type_str = {
    "string": "str",
    "integer": "int",
    "number": "float",
    "boolean": "bool",
}

_spec_api_type_to_python_type = {
    "string": str,
    "integer": int,
    "number": float,
    "boolean": bool,
}


def _create_run_artifacts_dir(action: "Action", run_id: str) -> str:
    """
    Returns:
        The path, relative to the settings.artifacts_dir which should be used
        to store the output of the given run.
    """
    from ._settings import get_settings

    settings = get_settings()
    artifacts_dir = settings.artifacts_dir
    path = run_id
    target_dir = artifacts_dir / path
    target_dir.mkdir(parents=True, exist_ok=False)
    return path


def _create_run(
    action: "Action", run_id: str, inputs: dict, relative_artifacts_dir: str
) -> "Run":
    from robocorp.action_server._models import RUN_ID_COUNTER, Counter

    from ._database import datetime_to_str
    from ._models import Run, RunStatus, get_db
    from ._runs_state_cache import get_global_runs_state

    db = get_db()
    run_kwargs: Dict[str, Any] = dict(
        id=run_id,
        status=RunStatus.NOT_RUN,
        action_id=action.id,
        start_time=datetime_to_str(datetime.datetime.now(datetime.timezone.utc)),
        run_time=None,
        inputs=json.dumps(inputs),
        result=None,
        error_message=None,
        relative_artifacts_dir=relative_artifacts_dir,
    )
    with db.transaction():
        with db.cursor() as cursor:
            db.execute_update_returning(
                cursor,
                "UPDATE counter SET value=value+1 WHERE id=? RETURNING value",
                [RUN_ID_COUNTER],
            )
            counter_record = cursor.fetchall()
            if not counter_record:
                raise RuntimeError(
                    f"Error. No counter found for run_id. Counters in db: {db.all(Counter)}"
                )
            run_kwargs["numbered_id"] = counter_record[0][0]

        run = Run(**run_kwargs)
        db.insert(run)

    # Ok, transaction finished properly. Let's add it to our in-memory cache.
    global_runs_state = get_global_runs_state()
    global_runs_state.on_run_inserted(run)

    return run


def _update_run(run: "Run", initial_time: float, run_finished: bool, **changes):
    from ._models import get_db
    from ._runs_state_cache import get_global_runs_state

    if run_finished:
        changes["run_time"] = time.monotonic() - initial_time

    db = get_db()
    for k, v in changes.items():
        setattr(run, k, v)
    fields_changed = tuple(changes.keys())
    with db.transaction():
        db.update(run, *fields_changed)

    # Ok, transaction finished properly. Let's update our in-memory cache.
    global_runs_state = get_global_runs_state()
    global_runs_state.on_run_changed(run, changes)


def _set_run_as_finished_ok(run: "Run", result: str, initial_time: float):
    from ._models import RunStatus

    _update_run(run, initial_time, True, result=result, status=RunStatus.PASSED)


def _set_run_as_finished_failed(run: "Run", error_message: str, initial_time: float):
    from ._models import RunStatus

    _update_run(
        run, initial_time, True, status=RunStatus.FAILED, error_message=error_message
    )


def _set_run_as_running(run: "Run", initial_time: float):
    from ._models import RunStatus

    _update_run(run, initial_time, False, status=RunStatus.RUNNING)


def _run_action_in_thread(
    action: "Action",
    signature: inspect.Signature,
    headers: Dict[str, str],
    *args,
    **kwargs,
):
    """
    This is where the user actually runs something.

    This runs in a thread (so, be careful when talking to the database).

    We have to take care of making a run with the proper environment,
    creating the run, collecting output info, etc.
    """
    from robocorp.action_server._gen_ids import gen_uuid
    from robocorp.action_server._settings import get_settings

    from ._actions_process_pool import get_actions_process_pool
    from ._models import Run, get_db

    settings = get_settings()
    ret_type = kwargs.pop("__ret_type__", "")
    if not ret_type:
        ret_type = "string"

    db = get_db()
    with db.connect():  # Connection is per-thread, so, we need to create a new one.
        inputs = {}
        if args or kwargs:
            bound = signature.bind(*args, **kwargs)

            for k, v in bound.arguments.items():
                inputs[k] = v

        expected_return_type = _spec_api_type_to_python_type.get(ret_type)
        if not expected_return_type:
            raise RuntimeError(
                f"The return type in the spec for the action {action.name} is not valid. Found: {ret_type}."
            )

        actions_process_pool = get_actions_process_pool()
        with actions_process_pool.obtain_process_for_action(action) as process_handle:
            run_id = gen_uuid("run")
            relative_artifacts_path: str = _create_run_artifacts_dir(action, run_id)
            run: Run = _create_run(action, run_id, inputs, relative_artifacts_path)

            input_json = (
                settings.artifacts_dir
                / relative_artifacts_path
                / "__action_server_inputs.json"
            )
            input_json.write_text(json.dumps(inputs))

            robot_artifacts = settings.artifacts_dir / relative_artifacts_path

            result_json = (
                settings.artifacts_dir
                / relative_artifacts_path
                / "__action_server_result.json"
            )

            output_file = (
                settings.artifacts_dir
                / relative_artifacts_path
                / "__action_server_output.txt"
            )

            initial_time = time.monotonic()
            try:
                _set_run_as_running(run, initial_time)
                returncode = process_handle.run_action(
                    action,
                    input_json,
                    robot_artifacts,
                    output_file,
                    result_json,
                    headers,
                )

                if returncode == 0:
                    try:
                        result: str = result_json.read_text("utf-8", "replace")
                    except Exception:
                        raise RuntimeError(
                            "It was not possible to collect the contents of the "
                            "result (json not created)."
                        )
                    try:
                        _set_run_as_finished_ok(run, result, initial_time)
                        ret = json.loads(result)
                    except Exception:
                        raise RuntimeError(
                            "Error loading the contents of {result_json} as json."
                        )
                    else:
                        if not isinstance(ret, expected_return_type):
                            raise RuntimeError(
                                f"Error in action. Expected return type: {ret_type}. Found return type: {type(ret)} (value: {ret})."
                            )
                        return ret

                raise RuntimeError(
                    f"Error: the process did not complete successfully when "
                    f"running. returncode: {returncode}"
                )

            except BaseException as e:
                _set_run_as_finished_failed(run, str(e), initial_time)
                raise


def _name_as_class_name(name):
    return name.replace("_", " ").title().replace(" ", "")


def generate_func_from_action(action: "Action"):
    """
    This function generates a method from the given action.

    def method(
        name: Annotated[str, Param(description="This is the name")],
        title: Annotated[str, Param(description="This is the title")] = None,
    ) -> int:
        pass

    class MethodInput(BaseModel):
        name: Annotated[str, Param(description="This is the name")]
        title: Annotated[str, Param(description="This is the title")] = None

    def method(
        args: MethodInput
    ) -> int:
        return 1
    """
    input_schema_dict = json.loads(action.input_schema)
    output_schema_dict = json.loads(action.output_schema)
    properties = input_schema_dict.get("properties", {})
    arguments = []
    argument_names = []
    for param_name, param_data in properties.items():
        desc = param_data.get("description", "")
        argument_names.append(f"args.{param_name}")
        param_type = param_data.get("type", "string")

        argument = f"""
    {param_name}: Annotated[{_spec_api_type_to_python_type_str[param_type]}, 
        Param(description={desc!r})]"""
        if "default" in param_data:
            default = param_data.get("default")
            argument = f"{argument}={default!r}"
        arguments.append(argument)

    # Note: Headers are explicitly hidden from spec to make the OpenAPI schema compatible with OpenAI
    headers = {
        "x_action_trace": "Optional[str] = Header(None, description='Client application run trace reference', alias='X-action-trace', include_in_schema=False)"
    }
    headers_as_params = [f"{key}: {value}" for key, value in headers.items()]
    headers_as_values = [f"'{key}': {key}" for key, _ in headers.items()]

    ret_type = output_schema_dict.get("type", "")
    ret = ""
    if ret_type:
        ret_desc = output_schema_dict.get("description")
        ret = f"""-> Annotated[{_spec_api_type_to_python_type_str[ret_type]}, 
        Param(description={ret_desc!r})]"""
    argument_names.append(f"__ret_type__={ret_type!r}")

    code = f"""
class {_name_as_class_name(action.name)}Input(BaseModel):
    pass
{''.join(arguments)}
    
def {action.name}_as_params({', '.join(arguments)}){ret}:
    pass

def {action.name}(args:{_name_as_class_name(action.name)}Input, {", ".join(headers_as_params)}){ret}:
    headers = {{{", ".join(headers_as_values)}}}
    return _run_action_in_thread(action, signature, headers, {', '.join(argument_names)})

signature = inspect.signature({action.name}_as_params)
"""

    compiled = compile(code, "<string>", "exec")
    ctx: dict = {
        "Annotated": Annotated,
        "Param": Param,
        "Header": Header,
        "BaseModel": BaseModel,
        "Optional": Optional,
        "action": action,
        "inspect": inspect,
        "_run_action_in_thread": _run_action_in_thread,
    }
    exec(compiled, ctx)
    method = ctx[action.name]
    return method



from __future__ import annotations

import itertools
import logging
import os
import subprocess
import sys
import threading
from concurrent.futures import Future
from functools import partial
from pathlib import Path
from typing import Any, Dict, List, Optional, Protocol, Union

from .callback import Callback

PathLike = Union[str, Path]

log = logging.getLogger(__name__)
IS_WINDOWS = sys.platform == "win32"


def _run_in_thread_and_get_result(target) -> Any:
    fut: Future[Any] = Future()

    def method():
        try:
            fut.set_result(target())
        except BaseException as e:
            fut.set_exception(e)

    t = threading.Thread(target=method)
    t.start()
    return fut.result()


def _stream_reader(stream, on_output):
    try:
        for line in iter(stream.readline, b""):
            if not line:
                break
            try:
                on_output(line)
            except Exception as e:
                log.error("Error reporting output line: %s", e)
    except Exception as e:
        log.error("Error reading stream: %s", e)


def _start_reader_threads(process, on_stdout, on_stderr):
    threads = []
    if on_stdout is not None:
        threads.append(
            threading.Thread(
                target=_stream_reader,
                args=(process.stdout, on_stdout),
                name="stream_reader_stdout",
            )
        )

    if on_stderr is not None:
        threads.append(
            threading.Thread(
                target=_stream_reader,
                args=(process.stderr, on_stderr),
                name="stream_reader_stderr",
            )
        )

    for t in threads:
        t.start()


class Process:
    _UID = itertools.count(0)

    def __init__(
        self,
        args: List[str],
        cwd: Optional[PathLike] = None,
        env: Optional[Dict[str, str]] = None,
    ):
        self._args = args
        self._uid = next(self._UID)
        self._cwd = cwd or Path.cwd()
        self._env = {**os.environ, **(env or {})}
        self._proc: Optional[subprocess.Popen] = None
        self.on_stderr = Callback()
        self.on_stdout = Callback()

    def is_alive(self):
        if self._proc is None:
            raise RuntimeError(
                "Process is still None (start() not called before is_alive)."
            )

        return is_process_alive(self._proc.pid)

    @property
    def returncode(self):
        if self._proc is None:
            raise RuntimeError(
                "Process is still None (start() not called before returncode)."
            )
        return self._proc.poll()

    def join(self):
        if self._proc is None:
            raise RuntimeError(
                "Process is still None (start() not called before join)."
            )
        self._proc.wait()

    def start(self, read_stderr: bool = True, read_stdout: bool = True) -> None:
        kwargs = build_subprocess_kwargs(
            self._cwd, self._env, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        log.debug(
            "Subprocess start [args=%s,cwd=%s,uid=%d]", self._args, self._cwd, self._uid
        )
        proc = self._proc = _run_in_thread_and_get_result(
            partial(_popen_raise, self._args, **kwargs)
        )
        log.debug("Subprocess started [pid=%s,uid=%d]", proc.pid, self._uid)
        on_stdout = None
        if read_stdout:
            on_stdout = self._on_stdout

        on_stderr = None
        if read_stderr:
            on_stderr = self._on_stderr
        _start_reader_threads(self._proc, on_stdout, on_stderr)

    def _on_stderr(self, line):
        if len(self.on_stderr) > 0:
            self.on_stderr(line.decode("utf-8", "replace"))

    def _on_stdout(self, line):
        if len(self.on_stdout) > 0:
            self.on_stdout(line.decode("utf-8", "replace"))

    def stop(self) -> None:
        if not self._proc:
            return
        if self._proc.poll() is not None:
            return
        if not self.is_alive():
            return

        log.debug("Subprocess kill [pid=%s,uid=%d]", self._proc.pid, self._uid)
        kill_process_and_subprocesses(self._proc.pid)

    @property
    def pid(self):
        return self._proc.pid


def build_subprocess_kwargs(cwd, env, **kwargs) -> dict:
    startupinfo = None
    if sys.platform == "win32":
        # We don't want to show the shell on windows!
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        startupinfo.wShowWindow = subprocess.SW_HIDE
        startupinfo = startupinfo

    if cwd:
        kwargs["cwd"] = cwd
    if env:
        kwargs["env"] = env
    kwargs["startupinfo"] = startupinfo
    return kwargs


def build_python_launch_env(new_env_vars: dict[str, str]) -> dict[str, str]:
    """
    Args:
        new_env_vars: If empty this means that this is an unmanaged env. In this
            case the environment used will be the same one used to run the
            action server itself.
    """
    if sys.platform == "win32":
        env = {}
        for k, v in os.environ.items():
            env[k.upper()] = v
    else:
        env = dict(os.environ)

    if new_env_vars:
        env.pop("PYTHONPATH", "")
        env.pop("PYTHONHOME", "")
        env.pop("VIRTUAL_ENV", "")

    env["PYTHONIOENCODING"] = "utf-8"
    env["PYTHONUNBUFFERED"] = "1"
    env.update(new_env_vars)

    # for k, v in env.items():
    #     if k in new_env_vars:
    #         print("FROM RCC", k, "=", v)
    # for k, v in env.items():
    #     if k not in new_env_vars:
    #         print("FROM SYS", k, "=", v)

    return env


if sys.platform == "win32":
    import ctypes

    kernel32 = ctypes.windll.kernel32
    PROCESS_SYNCHRONIZE = 0x00100000
    DWORD = ctypes.c_uint32
    BOOL = ctypes.c_int
    LPVOID = ctypes.c_void_p
    HANDLE = LPVOID

    OpenProcess = kernel32.OpenProcess
    OpenProcess.argtypes = [DWORD, BOOL, DWORD]
    OpenProcess.restype = HANDLE

    WaitForSingleObject = kernel32.WaitForSingleObject
    WaitForSingleObject.argtypes = [HANDLE, DWORD]
    WaitForSingleObject.restype = DWORD

    WAIT_TIMEOUT = 0x00000102
    WAIT_ABANDONED = 0x00000080
    WAIT_OBJECT_0 = 0
    WAIT_FAILED = 0xFFFFFFFF

    def is_process_alive(pid):
        """Check whether the process with the given pid is still alive.

        Running `os.kill()` on Windows always exits the process, so it can't be used to
        check for an alive process.
        see: https://docs.python.org/3/library/os.html?highlight=os%20kill#os.kill

        Hence ctypes is used to check for the process directly via windows API avoiding
        any other 3rd-party dependency.

        Args:
            pid (int): process ID

        Returns:
            bool: False if the process is not alive or don't have permission to check,
            True otherwise.
        """
        process = OpenProcess(PROCESS_SYNCHRONIZE, 0, pid)
        if process != 0:
            try:
                wait_result = WaitForSingleObject(process, 0)
                if wait_result == WAIT_TIMEOUT:
                    return True
            finally:
                kernel32.CloseHandle(process)
        return False

else:
    import errno

    def _is_process_alive(pid):
        """Check whether the process with the given pid is still alive.

        Args:
            pid (int): process ID

        Returns:
            bool: False if the process is not alive or don't have permission to check,
            True otherwise.
        """
        if pid < 0:
            return False
        try:
            os.kill(pid, 0)
        except OSError as e:
            if e.errno == errno.ESRCH:
                return False  # No such process.
            elif e.errno == errno.EPERM:
                return True  # permission denied.
            else:
                log.info("Unexpected errno: %s", e.errno)
                return False
        else:
            return True

    def is_process_alive(pid):
        if _is_process_alive(pid):
            # Check if zombie...
            try:
                cmd = ["ps", "-p", str(pid), "-o", "stat"]
                try:
                    process = subprocess.Popen(
                        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                    )
                except Exception:
                    log.exception("Error calling: %s.", " ".join(cmd))
                else:
                    stdout, _ = process.communicate()
                    stdout = stdout.decode("utf-8", "replace")
                    lines = [line.strip() for line in stdout.splitlines()]
                    if len(lines) > 1:
                        if lines[1].startswith("Z"):
                            return False  # It's a zombie
            except Exception:
                log.exception("Error checking if process is alive.")

            return True
        return False


def _popen(cmdline, **kwargs):
    try:
        popen = subprocess.Popen(cmdline, **kwargs)
        # Not sure why, but (just when running in VSCode) something as:
        # launching sys.executable actually got stuck unless a \n was written
        # (even if stdin was closed it wasn't enough).
        # -- note: this may be particular to my machine (fabioz), but it
        # may also be related to VSCode + Windows 11 + Windows Defender + python
        _stdin_write(popen, b"\n")
        return popen
    except Exception:
        log.exception("Error running: %s", (" ".join(cmdline)))
        return None


def _popen_raise(cmdline, **kwargs):
    try:
        popen = subprocess.Popen(cmdline, **kwargs)
        # Not sure why, but (just when running in VSCode) something as:
        # launching sys.executable actually got stuck unless a \n was written
        # (even if stdin was closed it wasn't enough).
        # -- note: this may be particular to my machine (fabioz), but it
        # may also be related to VSCode + Windows 11 + Windows Defender + python
        _stdin_write(popen, b"\n")
        return popen
    except Exception:
        log.exception("Error running: %s", (" ".join(cmdline)))
        raise


def _stdin_write(process, input):
    if process.stdin is not None:
        if input:
            try:
                process.stdin.write(input)
            except BrokenPipeError:
                pass  # communicate() must ignore broken pipe errors.
            except OSError as exc:
                if exc.errno == errno.EINVAL:
                    # bpo-19612, bpo-30418: On Windows, stdin.write() fails
                    # with EINVAL if the child process exited or if the child
                    # process is still running but closed the pipe.
                    pass
                else:
                    raise

        try:
            process.stdin.close()
        except BrokenPipeError:
            pass  # communicate() must ignore broken pipe errors.
        except OSError as exc:
            if exc.errno == errno.EINVAL:
                pass
            else:
                raise


def _call(cmdline, **kwargs):
    try:
        subprocess.check_call(cmdline, **kwargs)
    except Exception:
        log.exception("Error running: %s", (" ".join(cmdline)))
        return None


def _kill_process_and_subprocess_linux(pid):
    initial_pid = pid

    def list_children_and_stop_forking(ppid):
        children_pids = []
        _call(
            ["kill", "-STOP", str(ppid)], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        list_popen = _popen(
            ["pgrep", "-P", str(ppid)], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        if list_popen is not None:
            stdout, _ = list_popen.communicate()
            for line in stdout.splitlines():
                line = line.decode("ascii").strip()
                if line:
                    pid = str(line)
                    children_pids.append(pid)
                    # Recursively get children.
                    children_pids.extend(list_children_and_stop_forking(pid))
        return children_pids

    previously_found = set()

    for _ in range(50):  # Try this at most 50 times before giving up.
        children_pids = list_children_and_stop_forking(initial_pid)
        found_new = False

        for pid in children_pids:
            if pid not in previously_found:
                found_new = True
                previously_found.add(pid)
                _call(
                    ["kill", "-KILL", str(pid)],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )

        if not found_new:
            break

    # Now, finish the initial one.
    _call(
        ["kill", "-KILL", str(initial_pid)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )


def kill_process_and_subprocesses(pid):
    log.debug("Killing process and subprocesses of: %s", pid)
    from subprocess import CalledProcessError

    if sys.platform == "win32":
        args = ["taskkill", "/F", "/PID", str(pid), "/T"]
        retcode = subprocess.call(
            args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE
        )
        if retcode not in (0, 128, 255):
            raise CalledProcessError(retcode, args)
    else:
        _kill_process_and_subprocess_linux(pid)


class IProgressReporter(Protocol):
    @property
    def cancelled(self) -> bool:
        pass

    def set_additional_info(self, additional_info: str) -> None:
        """
        The progress reporter shows the title and elapsed time automatically.

        With this API it's possible to add additional info for the user to see.
        """


def check_output_interactive(
    *popenargs,
    timeout=None,
    progress_reporter: Optional[IProgressReporter] = None,
    on_stderr=lambda *args, **kwargs: None,
    on_stdout=lambda *args, **kwargs: None,
    **kwargs,
) -> bytes:
    """
    This has the same API as subprocess.check_output, but allows us to work with
    the contents being generated by the subprocess before the subprocess actually
    finishes.

    :param on_stderr:
        A callable(string) (called from another thread) whenever output is
        written with stderr contents.

    :param on_stdout:
        A callable(string) (called from another thread) whenever output is
        written with stdout contents.

    :return: the stdout generated by the command.
    """

    if kwargs.get("stdout", subprocess.PIPE) != subprocess.PIPE:
        raise AssertionError("stdout must be subprocess.PIPE")

    if kwargs.get("stderr", subprocess.PIPE) != subprocess.PIPE:
        # We could potentially also accept `subprocess.STDOUT`, but let's leave
        # this as a future improvement for now.
        raise AssertionError("stderr must be subprocess.PIPE")

    kwargs["stdout"] = subprocess.PIPE
    kwargs["stderr"] = subprocess.PIPE

    stdout_contents: List[bytes] = []
    stderr_contents: List[bytes] = []

    def stream_reader(stream, callback, contents_list: List[bytes]):
        for line in iter(stream.readline, b""):
            if not line:
                break
            contents_list.append(line)
            callback(line)

    def check_progress_cancelled(process, progress_reporter: IProgressReporter):
        try:
            while process.poll() is None:
                try:
                    process.wait(1)
                except BaseException:
                    if progress_reporter.cancelled:
                        retcode = process.poll()
                        if retcode is None:
                            msg_str = f"Progress was cancelled (RCC pid: {process.pid} was killed).\n"
                            msg = msg_str.encode("utf-8")
                            log.info(msg_str)
                            stderr_contents.insert(0, msg)
                            stderr_contents.append(msg)
                            on_stderr(msg)
                            kill_process_and_subprocesses(process.pid)
        except BaseException:
            log.exception("Error checking that progress was cancelled.")

    with subprocess.Popen(*popenargs, **kwargs) as process:
        threads = [
            threading.Thread(
                target=stream_reader,
                args=(process.stdout, on_stdout, stdout_contents),
                name="stream_reader_stdout",
            ),
            threading.Thread(
                target=stream_reader,
                args=(process.stderr, on_stderr, stderr_contents),
                name="stream_reader_stderr",
            ),
        ]
        if progress_reporter is not None:
            t = threading.Thread(
                target=check_progress_cancelled,
                args=(process, progress_reporter),
                name="check_progress_cancelled",
            )
            t.start()

        for t in threads:
            t.start()

        retcode: Optional[int]
        try:
            try:
                retcode = process.wait(timeout)
            except BaseException:
                # i.e.: KeyboardInterrupt / TimeoutExpired
                retcode = process.poll()

                if retcode is None:
                    # It still hasn't completed: kill it.
                    try:
                        kill_process_and_subprocesses(process.pid)
                    except BaseException:
                        log.exception("Error killing pid: %s" % (process.pid,))

                    retcode = process.wait()
                raise

        finally:
            for t in threads:
                t.join()

        if retcode:
            raise subprocess.CalledProcessError(
                retcode,
                process.args,
                output=b"".join(stdout_contents),
                stderr=b"".join(stderr_contents),
            )

        return b"".join(stdout_contents)


_track_pids_to_exit = set()
_watching_thread_global = None
PARENT_PROCESS_WATCH_INTERVAL = 1


def exit_when_pid_exists(pid, interval=PARENT_PROCESS_WATCH_INTERVAL):
    _track_pids_to_exit.add(pid)
    global _watching_thread_global
    if _watching_thread_global is None:
        import time

        def watch_parent_process():
            # exit when any of the ids we're tracking exit.
            while True:
                for pid in _track_pids_to_exit:
                    if not is_process_alive(pid):
                        # Note: just exit since the parent process already
                        # exited.
                        log.debug(
                            f"Force-quit process: {os.getpid()} because parent: {pid} exited"
                        )
                        os._exit(0)

                time.sleep(interval)

        _watching_thread_global = threading.Thread(target=watch_parent_process, args=())
        _watching_thread_global.daemon = True
        _watching_thread_global.start()



"""
The way that migrations work is the following:

We have a table with the id and name of each migration applied.

When starting up we connect to the db and see the current version.
Then, we apply each pending migration since the last version.
"""

import logging
import os.path
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Union

log = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from robocorp.action_server._database import Database

MIGRATION_ID_TO_NAME: Dict[int, str] = {
    # we'll look for a 'migration_initial' module based on this.
    1: "initial",
    # we'll look for a 'migration_add_action_enabled' module based on this.
    2: "add_action_enabled",
    # we'll look for a 'migration_add_is_consequential' module based on this.
    3: "add_is_consequential",
}

CURRENT_VERSION: int = max(MIGRATION_ID_TO_NAME.keys())


@dataclass
class Migration:
    id: int  # Migration id
    name: str  # Migration name


def _migrate_to(db: "Database", db_migration_version: int) -> None:
    from importlib import import_module

    name = MIGRATION_ID_TO_NAME[db_migration_version]
    mod = import_module(f"robocorp.action_server.migrations.migration_{name}")
    mod.migrate(db)


def migrate_db(
    db_path: Union[Path, str],
    to_version=CURRENT_VERSION,
    database: Optional["Database"] = None,
) -> bool:
    """
    Returns true if the migration worked properly or if the migration was not needed
    and false if the migration could not be done for some reason.

    :param database: Expected to be passed when dealing with an in-memory database.
    """
    assert os.path.exists(db_path), f"Unable to do migration. {db_path} does not exist."

    if not db_migration_pending(db_path):
        return True  # It's already correct

    import shutil
    import time

    path = Path(db_path)
    # Ok, we need to do a migration. The first thing is creating a backup,
    # just in case something goes bad.
    parent_dir = path.parent
    name = path.name
    log.info("Preparing to migrate database at: %s", db_path)
    backup_file = parent_dir / f"{name}-pre-migration-{to_version}-{time.time()}.bak"
    log.info("Creating backup at: %s", backup_file)

    shutil.copyfile(path, backup_file)

    from robocorp.action_server._database import Database
    from robocorp.action_server._models import get_all_model_classes

    db = Database(db_path)
    with db.connect():
        with db.transaction():
            db.initialize(get_all_model_classes())
            if "migration" not in db.list_table_names():
                raise RuntimeError(
                    f"""Error: 
It seems that this version of the database ({db.db_path}) is too old.
Please erase it and recreate it from scratch."""
                )
            else:
                migrations = db.all(Migration)
                if not migrations:
                    raise RuntimeError(
                        f"""Error: 
It seems that this version of the database ({db.db_path}) is too old.
Please erase it and recreate it from scratch."""
                    )

                else:
                    db_migration_version = max(x.id for x in migrations)

            while db_migration_version < to_version:
                db_migration_version += 1
                log.info(
                    "Will migrate to version: %s (%s)",
                    db_migration_version,
                    MIGRATION_ID_TO_NAME[db_migration_version],
                )
                _migrate_to(db, db_migration_version)

    return True


def _db_migration_pending(db: "Database") -> bool:
    if "migration" not in db.list_table_names():
        return True

    db.initialize([Migration])
    migrations = db.all(Migration)
    if not migrations:
        # No migrations applied, do it now.
        return True

    latest_migration_applied = max(x.id for x in migrations)

    if latest_migration_applied == CURRENT_VERSION:
        return False

    return True


def db_migration_pending(db_path: Union[Path, str]) -> bool:
    from robocorp.action_server._database import Database

    if db_path == ":memory:":
        raise RuntimeError("Migration support not available for in-memory database.")

    path = Path(db_path)
    if not path.exists():
        raise RuntimeError(
            f"Unable to check if migration is pending because file: {db_path} does not exist."
        )

    # Ok, it already exists. Check if it's pending a migration.
    db = Database(db_path)
    with db.connect():
        return _db_migration_pending(db)



# Original work Copyright 2018 Palantir Technologies, Inc. (MIT)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import queue
import threading
from typing import Optional

log = logging.getLogger(__name__)


def read(stream) -> Optional[str]:
    """
    Reads one message from the stream and returns the message (or None if EOF was reached).

    :param stream:
        The stream we should be reading from.

    :return str|NoneType:
        The message or None if the stream was closed.
    """
    headers = {}
    while True:
        # Interpret the http protocol headers
        line = stream.readline()  # The trailing \r\n should be there.

        if not line:  # EOF
            return None
        line = line.strip().decode("ascii")
        if not line:  # Read just a new line without any contents
            break
        try:
            name, value = line.split(": ", 1)
        except ValueError:
            raise RuntimeError("Invalid header line: {}.".format(line))
        headers[name.strip()] = value.strip()

    if not headers:
        raise RuntimeError("Got message without headers.")

    content_length = int(headers["Content-Length"])

    # Get the actual json
    body = _read_len(stream, content_length)
    return body.decode("utf-8")


def _read_len(stream, content_length) -> bytes:
    buf = b""
    if not content_length:
        return buf

    # Grab the body
    while True:
        data = stream.read(content_length - len(buf))
        if not buf and len(data) == content_length:
            # Common case
            return data
        buf += data
        if len(buf) == content_length:
            return buf
        if len(buf) > content_length:
            raise AssertionError(
                "Expected to read message up to len == %s (already read: %s). Found:\n%s"
                % (content_length, len(buf), buf.decode("utf-8", "replace"))
            )
        # len(buf) < content_length (just keep on going).


class JsonRpcStreamReaderThread(threading.Thread):
    def __init__(self, rfile, queue, message_consumer):
        threading.Thread.__init__(self)
        self._rfile = rfile
        self._queue = queue
        self._message_consumer = message_consumer
        self.name = "JsonRpcStreamReaderThread"
        self.daemon = True

    def run(self):
        try:
            while not self._rfile.closed:
                data = read(self._rfile)
                if data is None:
                    log.debug("Read: %s", data)
                    return

                try:
                    msg = json.loads(data)
                except:
                    log.exception("Failed to parse JSON message %s", data)
                    continue

                if isinstance(msg, dict):
                    # Note: parsing is done on a thread so that we can read
                    # while processing so that we can give priority to `cancelProgress`.

                    if msg.get("method") == "cancelProgress":
                        try:
                            self._message_consumer(msg)
                        except:
                            log.exception("Error processing JSON message %s", msg)
                        continue

                    log.debug("Read: %s", data)
                else:
                    log.debug("Read (non dict data): %s", data)

                self._queue.put(msg)

                if isinstance(msg, dict):
                    # When we receive an exit we stop reading.
                    if msg.get("method") == "exit":
                        return
        except ConnectionResetError:
            pass  # Just ignore this one (connection was closed)
        except Exception:
            log.exception("Error in JsonRpcStreamReader.")
        finally:
            self._queue.put(None)

    @property
    def _name(self):
        return f"{self.__name} (closed: {self._rfile.closed})"

    @_name.setter
    def _name(self, name):
        self.__name = name


class JsonRpcStreamReader(object):
    def __init__(self, rfile):
        self._rfile = rfile
        self._queue = queue.Queue()
        self._reader_thread = None

    def get_read_queue(self):
        return self._queue

    def close(self):
        pass
        # We don't close the reader because it can deadlock if someone
        # is currently reading.
        # self._rfile.close()

    def listen(self, message_consumer):
        """Blocking call to listen for messages on the rfile.

        Args:
            message_consumer (fn): function that is passed each message as it is read off the socket.
        """
        self._reader_thread = JsonRpcStreamReaderThread(
            self._rfile, self._queue, message_consumer
        )
        self._reader_thread.start()
        try:
            while True:
                msg = self._queue.get()
                if msg is None:
                    break

                try:
                    if hasattr(msg, "__call__"):
                        # Clients can use get_read_queue().put(lambda: ...)
                        # to process something in the main thread with a callable.
                        msg()
                    else:
                        message_consumer(msg)
                except:
                    log.exception("Error processing JSON message %s", msg)
                    continue

        except ConnectionResetError:
            pass  # Just ignore this one (connection was closed)
        except Exception:
            log.exception("Error in JsonRpcStreamReader.")
        finally:
            log.debug("Exited JsonRpcStreamReader.")


class JsonRpcStreamWriter(object):
    def __init__(self, wfile, **json_dumps_args):
        assert wfile is not None
        self._wfile = wfile
        self._wfile_lock = threading.Lock()
        self._json_dumps_args = json_dumps_args

    def close(self):
        log.debug("Will close writer")
        with self._wfile_lock:
            self._wfile.close()

    def write(self, message):
        with self._wfile_lock:
            if self._wfile.closed:
                log.debug("Unable to write %s (file already closed).", (message,))
                return False
            try:
                if isinstance(message, dict):
                    log.debug("Writing: %s", message)
                else:
                    log.debug("Writing (non dict message): %s", message)

                body = json.dumps(message, **self._json_dumps_args)

                as_bytes = body.encode("utf-8")
                stream = self._wfile
                content_len_as_str = "Content-Length: %s\r\n\r\n" % len(as_bytes)
                content_len_bytes = content_len_as_str.encode("ascii")

                stream.write(content_len_bytes)
                stream.write(as_bytes)
                stream.flush()
                return True
            except Exception:  # pylint: disable=broad-except
                log.exception(
                    "Failed to write message to output file %s - closed: %s",
                    message,
                    self._wfile.closed,
                )
                return False



import logging
import os
import sys
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, Optional

log = logging.getLogger(__name__)


def is_frozen():
    if getattr(sys, "frozen", False):
        return True
    try:
        __file__
    except NameError:
        return True

    return False


def get_python_exe_from_env(env):
    python = env.get("PYTHON_EXE")
    if not python:
        if is_frozen():
            raise RuntimeError(
                "Unable to run because no 'conda.yaml' was present to bootstrap the environment\n"
                "(note: when the action server is distributed without sources, a 'conda.yaml' for "
                "the target environment is always required)."
            )
        else:
            python = sys.executable

    return python


def get_default_settings_dir() -> Path:
    if sys.platform == "win32":
        localappdata = os.environ.get("LOCALAPPDATA")
        if not localappdata:
            raise RuntimeError("Error. LOCALAPPDATA not defined in environment!")
        path = Path(localappdata) / "robocorp" / ".action_server"
    else:
        # Linux/Mac
        path = Path("~/robocorp/.action_server").expanduser()

    path.mkdir(parents=True, exist_ok=True)
    return path


@dataclass(slots=True, kw_only=True)
class Settings:
    artifacts_dir: Path
    datadir: Path

    title: str = "Robocorp Actions Server"

    address: str = "localhost"
    port: int = 8080
    verbose: bool = False
    db_file: str = "server.db"
    expose_url: str = "robocorp.link"
    server_url: str = "http://localhost:8080"

    min_processes: int = 0
    max_processes: int = 20
    reuse_processes: bool = False

    @classmethod
    def defaults(cls):
        fields = cls.__dataclass_fields__
        ret = {}
        SENTINEL = []
        for name, field in fields.items():
            v = getattr(field, "default", SENTINEL)
            if v is not SENTINEL:
                ret[name] = v
        return ret

    @classmethod
    def create(cls, args) -> "Settings":
        user_specified_datadir = args.datadir
        if not user_specified_datadir:
            import hashlib

            curr_cwd_dir = Path(".").absolute()
            name = curr_cwd_dir.name
            as_posix = curr_cwd_dir.as_posix()
            if sys.platform == "win32":
                as_posix = as_posix.lower()

            # Not secure, but ok for our purposes
            short_hash = hashlib.sha256(as_posix.encode()).hexdigest()[:8]
            datadir_name = f"{get_default_settings_dir()}/{name}_{short_hash}"

            log.info(f"Using datadir (scoped to the current directory): {datadir_name}")
            user_expanded_datadir = Path(datadir_name).expanduser()

        else:
            log.info(f"Using user-specified datadir: {user_specified_datadir}")
            user_expanded_datadir = Path(user_specified_datadir).expanduser()

        datadir = user_expanded_datadir.absolute()

        settings = Settings(datadir=datadir, artifacts_dir=datadir / "artifacts")
        # Optional (just in 'start' command, not in 'import')
        for attr in (
            "address",
            "port",
            "min_processes",
            "max_processes",
            "reuse_processes",
        ):
            assert hasattr(settings, attr)
            if hasattr(args, attr):
                setattr(settings, attr, getattr(args, attr))

        if hasattr(args, "server_url") and args.server_url is not None:
            settings.server_url = args.server_url
        else:
            settings.server_url = f"http://{settings.address}:{settings.port}"

        # Used in either import or start commands.
        settings.verbose = args.verbose
        settings.db_file = args.db_file
        return settings

    def to_uvicorn(self):
        return {
            "host": self.address,
            "port": self.port,
            "reload": False,
            "log_config": None,
        }


_global_settings: Optional[Settings] = None


@contextmanager
def setup_settings(args) -> Iterator[Settings]:
    global _global_settings
    settings = Settings.create(args)
    _global_settings = settings
    try:
        yield settings
    finally:
        _global_settings = None


def get_settings() -> Settings:
    if _global_settings is None:
        raise AssertionError("It seems that the settings have not been setup yet.")
    return _global_settings



import os
import typing
from pathlib import Path
from typing import Dict

if typing.TYPE_CHECKING:
    from robocorp.action_server._models import Action, ActionPackage
    from robocorp.action_server._settings import Settings


def _add_preload_actions_dir_to_env_pythonpath(env: Dict[str, str]) -> None:
    from robocorp.action_server import _preload_actions

    p = Path(_preload_actions.__file__)
    if "__init__" in p.name:
        p = p.parent
    curr_pythonpath = env.get("PYTHONPATH", "")
    if not curr_pythonpath:
        env["PYTHONPATH"] = str(p)
    else:
        env["PYTHONPATH"] = f"{p}{os.pathsep}{curr_pythonpath}"


def get_action_package_cwd(
    settings: "Settings", action_package: "ActionPackage"
) -> Path:
    directory = Path(action_package.directory)

    # Make it absolute now!
    if not directory.is_absolute():
        directory = (settings.datadir / directory).absolute()

    if not directory.exists():
        raise RuntimeError(
            f"""Error. Unable to run because action package directory:
{directory} 
does not exist.
This usually happens because the action was moved to a different
location. To fix please move it back to the original location or
import it again from the new location.
"""
        )
    return directory



import os

# This has to be imported once and we're done.
# Note that this module may be imported multiple times (at it's imported once
# and then it's removed from sys.modules and then reimported in a new run)
# as such, things that should be only done once need to be in a separate module.

try:
    import preload_actions_teardown  # type: ignore
except ImportError:
    from . import preload_actions_teardown

from robocorp.log import html

X_ACTION_TRACE = os.environ.get("X_ACTION_TRACE", "")

if X_ACTION_TRACE:
    html(
        f'<p style="line-height:34px">External <a href="{X_ACTION_TRACE}"'
        'style="color:rgb(var(--color-content-accent))">Client Application Trace</a> reported</p>'
    )



import asyncio
import logging
import threading
import typing
from dataclasses import asdict
from functools import partial

from fastapi.routing import APIRouter
from starlette.websockets import WebSocket, WebSocketDisconnect

if typing.TYPE_CHECKING:
    from ._models import Run
    from ._runs_state_cache import RunChangeEvent

log = logging.getLogger(__name__)


websocket_api_router = APIRouter(prefix="/api/ws")


class RunNotificationsThread(threading.Thread):
    def __init__(self, loop, websocket):
        super().__init__()

        self.loop = loop
        self.websocket = websocket
        self._disposed = threading.Event()
        self.daemon = True

    def __call__(self, run_change_event: "RunChangeEvent"):
        asyncio.run_coroutine_threadsafe(
            self._report_change_event(run_change_event), self.loop
        )

    def run(self):
        # Collect the runs, send them and register for additional changes until
        # `dispose` is called.
        from robocorp.action_server._runs_state_cache import get_global_runs_state

        global_runs_state = get_global_runs_state()

        with global_runs_state.semaphore:
            runs = global_runs_state.get_current_run_state()
            asyncio.run_coroutine_threadsafe(self._report_runs(runs), self.loop)
            global_runs_state.register(self)

        try:
            self._disposed.wait()
        finally:
            with global_runs_state.semaphore:
                global_runs_state.unregister(self)

    async def _report_change_event(self, run_change_event: "RunChangeEvent"):
        """
        This is run in asyncio, not in this thread.
        """
        if self._disposed.is_set():
            return

        try:
            if run_change_event.ev == "added":
                await self.websocket.send_json(
                    {
                        "message_type": "run_added",
                        "run": asdict(run_change_event.run),
                    }
                )
            elif run_change_event.ev == "changed":
                await self.websocket.send_json(
                    {
                        "message_type": "run_changed",
                        "run_id": run_change_event.run.id,
                        "changes": run_change_event.changes,
                    }
                )
            else:
                log.critical(f"Unexpected run change event: {run_change_event}.")
        except Exception:
            log.exception("Error reporting change event to json.")
            raise

    async def send_json(self, data: list[dict] | dict):
        """
        This is run in asyncio, not in this thread.
        """
        if self._disposed.is_set():
            return

        await self.websocket.send_json(data)

    async def _report_runs(self, runs: list["Run"]):
        """
        This is run in asyncio, not in this thread.
        """
        await self.send_json(
            {
                "message_type": "runs_collected",
                "runs": [asdict(run) for run in runs],
            }
        )

    def dispose(self):
        """
        Disposes of this thread.
        """
        self._disposed.set()


def _list_actions_in_threadpool(run_notifications, message_id):
    from robocorp.action_server._api_action_package import list_action_packages

    try:
        action_packages = list_action_packages()
    except Exception:
        log.exception("Error collection action packages.")
        action_packages = []

    asyncio.run_coroutine_threadsafe(
        run_notifications.send_json(
            {
                "message_type": "response",
                "data": {
                    "message_id": message_id,
                    "result": [asdict(p) for p in action_packages],
                },
            }
        ),
        run_notifications.loop,
    )


@websocket_api_router.websocket("")
async def websocket_endpoint(websocket: WebSocket):
    from starlette.concurrency import run_in_threadpool

    loop = asyncio.get_running_loop()

    await websocket.accept()
    run_notifications = RunNotificationsThread(loop, websocket)
    try:
        while True:
            data = await websocket.receive_json()
            message_type = data.get("message_type")
            if message_type == "ping":
                await websocket.send_json({"message_type": "pong"})

            elif message_type == "request":
                # We received some request from the UI, let's handle it.
                try:
                    request_data = data.get("data")
                    method = request_data.get("method")
                    if method == "GET":
                        message_id = request_data.get("message_id")
                        url = request_data.get("url")
                        if url == "/api/actionPackages":
                            loop.create_task(
                                run_in_threadpool(
                                    partial(
                                        _list_actions_in_threadpool,
                                        run_notifications,
                                        message_id,
                                    )
                                )
                            )

                except Exception:
                    log.exception("Unexpected exception handling a request")

            elif message_type == "start_listen_run_events":
                # In this case we have to:
                # 1. Send the current runs to the client.
                # 2. When run changes are done, notify the client.
                # The thread should take care of this after it's started.
                if not run_notifications.is_alive():
                    run_notifications.start()
                else:
                    # If the client calls start_listen_run_events, should we send
                    # it again? Right now clients are not expected to do that, so
                    # just log it and ignore.
                    log.info('Ignoring additional call to "start_listen_run_events"')
    except WebSocketDisconnect:
        log.info("Client disconnected from websocket.")
    except Exception:
        log.exception("Unexpected exception from websocket.")
    finally:
        run_notifications.dispose()



from robocorp.action_server._database import Database
from robocorp.action_server.migrations import Migration


def migrate(db: Database) -> None:
    from robocorp.action_server.migrations import MIGRATION_ID_TO_NAME

    db.execute(
        """
ALTER TABLE action
ADD COLUMN is_consequential INTEGER;
"""
    )

    db.insert(Migration(id=3, name=MIGRATION_ID_TO_NAME[3]))



import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Dict, List, Optional

import fastapi
from fastapi.params import Param
from fastapi.routing import APIRouter
from starlette.responses import FileResponse

from robocorp.action_server._models import Run

log = logging.getLogger(__name__)
run_api_router = APIRouter(prefix="/api/runs")


@run_api_router.get("", response_model=list[Run])
def list_runs():
    from ._runs_state_cache import get_global_runs_state

    global_runs_state = get_global_runs_state()
    with global_runs_state.semaphore:
        return global_runs_state.get_current_run_state()


def get_run_by_id(run_id: str) -> Run:
    try:
        from ._runs_state_cache import get_global_runs_state

        global_runs_state = get_global_runs_state()
        with global_runs_state.semaphore:
            return global_runs_state.get_run_from_id(run_id)
    except KeyError as err:
        from fastapi.exceptions import HTTPException
        from starlette import status

        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Unknown run id: {run_id}",
        ) from err


@run_api_router.get("/{run_id}", response_model=Run)
def show_run(run_id: str = fastapi.Path(title="ID for run")):
    # Note that we want to use
    return get_run_by_id(run_id)


@dataclass
class ArtifactInfo:
    name: str
    size_in_bytes: int


def _scandir_recursive(path):
    for entry in os.scandir(path):
        if entry.is_dir(follow_symlinks=False):
            yield from _scandir_recursive(entry.path)
        else:
            yield entry


def _get_file_info_in_path(path: Path) -> List[ArtifactInfo]:
    file_names = []
    for entry in _scandir_recursive(path):
        try:
            if entry.is_file():
                size = entry.stat().st_size
                name = Path(entry.path).relative_to(path).as_posix()
                file_names.append(ArtifactInfo(name, size))
        except Exception:
            log.exception(f"Unable to get information from: {entry}.")
            continue
    return file_names


@run_api_router.get("/{run_id}/artifacts")
def get_run_artifacts(
    run_id: str = fastapi.Path(title="ID for run"),
) -> Annotated[
    List[ArtifactInfo],
    Param(
        description="""Provides a list with the artifacts available
for a given run (i.e.: [{'name': '__action_server_output.txt', 'size_in_bytes': 22}])
"""
    ),
]:
    from robocorp.action_server._settings import get_settings

    settings = get_settings()
    run = get_run_by_id(run_id)
    artifacts_dir = settings.artifacts_dir

    if not settings.artifacts_dir:
        log.critical(
            "Unable to get artifacts because the settings artifacts_dir is not defined."
        )
        return []

    artifacts_in = artifacts_dir / run.relative_artifacts_dir
    if not artifacts_in.exists():
        log.critical(
            f"Unable to get artifacts because the artifacts_dir ({artifacts_in}) "
            "does not exist."
        )
        return []

    return _get_file_info_in_path(artifacts_in)


@run_api_router.get("/{run_id}/log.html")
def get_run_log_html(run_id: str = fastapi.Path(title="ID for run")):
    from fastapi.exceptions import HTTPException

    artifacts_in = _get_artifacts_dir_for_run_id(run_id)
    if artifacts_in is None:
        raise HTTPException(
            status_code=404, detail="log.html not available for requested run."
        )
    log_html = artifacts_in / "log.html"
    if not os.path.exists(log_html):
        raise HTTPException(
            status_code=404, detail="log.html not available for requested run."
        )
    return FileResponse(log_html)


def _get_artifacts_dir_for_run_id(run_id: str) -> Optional[Path]:
    from robocorp.action_server._settings import get_settings

    settings = get_settings()
    run = get_run_by_id(run_id)
    artifacts_dir = settings.artifacts_dir

    if not settings.artifacts_dir:
        log.critical(
            "Unable to get artifacts because the settings artifacts_dir is not defined."
        )
        return None

    artifacts_in = (artifacts_dir / run.relative_artifacts_dir).absolute()
    if not artifacts_in.exists():
        log.critical(
            f"Unable to get artifacts because the artifacts_dir ({artifacts_in}) "
            "does not exist."
        )
        return None

    return artifacts_in


@run_api_router.get("/{run_id}/artifacts/text-content")
def get_run_artifact_text(
    run_id: str = fastapi.Path(title="ID for run"),
    artifact_names: Optional[List[str]] = fastapi.Query(
        default=None, title="Artifact names for which the content should be gotten."
    ),
    artifact_name_regexp: Optional[str] = fastapi.Query(
        default=None,
        title="A regexp to match artifact names.",
    ),
) -> Dict[str, str]:
    artifacts_in = _get_artifacts_dir_for_run_id(run_id)
    if artifacts_in is None:
        return {}
    if artifact_names is None:
        artifact_names = []

    checked = set()

    if artifact_name_regexp:
        import re

        pattern = re.compile(artifact_name_regexp)

        # We can't use glob directly because users would be able to get contents
        # out of the artifacts dir with that.
        for artifact_info in _get_file_info_in_path(artifacts_in):
            if pattern.match(artifact_info.name):
                artifact_names.append(artifact_info.name)

    ret: Dict[str, str] = {}
    for name in artifact_names:
        if name in checked:
            continue
        checked.add(name)

        f = (artifacts_in / name).absolute()
        if not f.exists():
            log.critical(f"Unable to get artifact because it does not exist: {f}")
            continue

        try:
            Path(f).relative_to(artifacts_in)
        except ValueError:
            log.critical(
                "Unable to get artifact (%s) because it does not point to a folder "
                "inside of the artifacts dir (%s).",
                f,
                artifacts_in,
            )
            continue

        try:
            ret[name] = f.read_text("utf-8", "replace")
        except Exception:
            log.critical("Unable to read artifact: %s as text.", f)

            continue

    return ret


@run_api_router.get("/{run_id}/artifacts/binary-content", response_class=FileResponse)
def get_run_artifact_binary(
    run_id: str = fastapi.Path(title="ID for run"),
    artifact_name: str = fastapi.Query(
        title="Artifact name for which the content should be gotten."
    ),
):
    from robocorp.action_server._settings import get_settings

    settings = get_settings()
    run = get_run_by_id(run_id)
    artifacts_dir = settings.artifacts_dir

    if not settings.artifacts_dir:
        log.critical(
            "Unable to get artifacts because the settings artifacts_dir is not defined."
        )
        return None

    artifacts_in = artifacts_dir / run.relative_artifacts_dir
    if not artifacts_in.exists():
        log.critical(
            f"Unable to get artifacts because the artifacts_dir ({artifacts_in}) "
            "does not exist."
        )
        return None

    f = (artifacts_in / artifact_name).absolute()
    if not f.exists():
        log.critical(f"Unable to get artifact because it does not exist: {f}")
        return None

    try:
        Path(f).relative_to(artifacts_in)
    except ValueError:
        log.critical(
            "Unable to get artifact (%s) because it does not point to a folder "
            "inside of the artifacts dir (%s).",
            f,
            artifacts_dir,
        )
        return None

    return FileResponse(f)



import datetime
import itertools
import logging
import re
import sqlite3
import sys
import threading
from contextlib import closing, contextmanager
from pathlib import Path
from types import NoneType
from typing import Any, Dict, Iterator, List, Optional, Set, Tuple, Type, TypeVar, Union

log = logging.getLogger(__name__)

_RE_FIRST_CAP = re.compile("(.)([A-Z][a-z]+)")
_RE_ALL_CAP = re.compile("([a-z0-9])([A-Z])")


T = TypeVar("T")


def _make_table_name(cls_name):
    s1 = _RE_FIRST_CAP.sub(r"\1_\2", cls_name)
    return _RE_ALL_CAP.sub(r"\1_\2", s1).lower()


# Helper functions since we must store the datetime as str.
# (no need to autoconvert all the time).


def datetime_to_str(val: datetime.datetime) -> str:
    return val.isoformat()


def str_to_datetime(val: str) -> datetime.datetime:
    return datetime.datetime.fromisoformat(val)


class DBError(Exception):
    pass


class DBRules:
    def __init__(self) -> None:
        # Fields which should have unique indexes in the format:
        # "Class.field_name"
        self.unique_indexes: Set[str] = set()

        # Fields which should have indexes in the format:
        # "Class.field_name"
        self.indexes: Set[str] = set()

        # Fields which are foreign keys in the format:
        # "Class.field_name"
        self.foreign_keys: Set[str] = set()


class Database:
    """
    Some notes:

        Connections must be per-thread.

        No 2 connections should be writing at the same time (ideally, use a
        single thread for writing).

        This class makes it so that there's only one connection per thread.
    """

    verbose = 0

    def __init__(self, db_path: Optional[Union[Path, str]] = None):
        self._cls_to_type_hint: Dict[type, dict] = {}
        if not db_path:
            from ._settings import get_settings

            self._db_path = get_settings().datadir / "server.db"
        else:
            self._db_path = Path(db_path)

        self._table_name_to_cls: Dict[str, type] = {}
        self._tlocal = threading.local()
        self._counter = itertools.count(0)
        self._write_lock = threading.RLock()
        self._classes: List[type] = []

    @property
    def db_path(self) -> Path:
        return self._db_path

    def _get_type_hints(self, cls) -> dict:
        try:
            return self._cls_to_type_hint[cls]
        except KeyError:
            from typing import get_type_hints

            ret = get_type_hints(cls)
            self._cls_to_type_hint[cls] = ret
            return ret

    def _iter_name_and_name_cls_fields(self, cls) -> Iterator[Tuple[str, type]]:
        yield from self._get_type_hints(cls).items()

    def _iter_name_fields(self, cls) -> Iterator[str]:
        yield from self._get_type_hints(cls).keys()

    @contextmanager
    def connect(self) -> Iterator[None]:
        try:
            conn = self._tlocal.conn
        except AttributeError:
            conn = None

        if conn is not None:
            yield
            return

        with closing(sqlite3.connect(self._db_path, isolation_level=None)) as conn:
            conn.execute("PRAGMA foreign_keys = ON")
            self._tlocal.conn = conn
            try:
                yield
            finally:
                self._tlocal.conn = None

    def _next_savepoint_name(self):
        return f"savepoint_{next(self._counter)}"

    @contextmanager
    def cursor(self) -> Iterator[sqlite3.Cursor]:
        """
        A cursor should be requested to do queries.
        """
        try:
            conn = self._tlocal.conn
        except AttributeError:
            conn = None
        if conn is None:
            raise RuntimeError(
                "Error. Cannot create a cursor without a connection in place."
            )

        with closing(conn.cursor()) as cur:
            yield cur

    def in_transaction(self) -> bool:
        """
        Returns:
            True if we're currently in a transaction and False otherwise.
        """
        try:
            in_transaction = self._tlocal.in_transaction
        except AttributeError:
            in_transaction = self._tlocal.in_transaction = 0

        return in_transaction > 0

    @contextmanager
    def transaction(self) -> Iterator[None]:
        """
        A transaction should be created when contents are about to be written.

        SQLite can't deal with multiple writers, so, we use a lock in Python
        which will prevent other threads from writing at the same time.
        """

        try:
            conn = self._tlocal.conn
        except AttributeError:
            conn = None
        if conn is None:
            raise DBError(
                "Unable to create a transaction because no connection is in place."
            )

        try:
            in_transaction = self._tlocal.in_transaction
        except AttributeError:
            in_transaction = self._tlocal.in_transaction = 0

        if in_transaction:
            # Nested transactions are not supported, so, don't start a new one
            # here, but we can still use savepoints.
            self._tlocal.in_transaction += 1
            savepoint_name = self._next_savepoint_name()
            self.execute(f"savepoint {savepoint_name};")
            try:
                yield
            except BaseException:
                self.execute(f"rollback to savepoint {savepoint_name};")
                raise
            finally:
                self._tlocal.in_transaction -= 1
            return

        assert (
            self._tlocal.in_transaction == 0
        ), "Error transaction nesting logic not correct!"

        self._tlocal.in_transaction += 1
        try:
            self.execute("BEGIN")
            yield
        except BaseException:
            log.exception("Error. Rolling back database")
            conn.rollback()
            raise
        else:
            conn.commit()
        finally:
            self._tlocal.in_transaction -= 1
            assert (
                self._tlocal.in_transaction == 0
            ), "Error transaction nesting logic not correct!"

    def update(self, instance, *fields: str):
        """
        Updates database values from some instance given its id.
        """
        fields_dict: dict[str, Any] = {}
        for field in fields:
            fields_dict[field] = getattr(instance, field)
        self.update_by_id(instance.__class__, instance.id, fields_dict)

    def update_by_id(self, cls: Type, id: Any, fields: dict[str, Any]):
        """
        Updates database values from some instance given its id.
        """
        table_name = _make_table_name(cls.__name__)
        set_fields = []
        values = []
        for name, value in fields.items():
            set_fields.append(f"{name}=?")
            values.append(value)

        values.append(id)
        sql = f"UPDATE {table_name} SET {', '.join(set_fields)} WHERE id=?"
        self.execute(sql, values)

    def insert(
        self,
        instance,
    ) -> None:
        table_name = _make_table_name(instance.__class__.__name__)

        field_names = []
        values = []
        placeholders = []
        for name in self._iter_name_fields(instance.__class__):
            field_names.append(name)
            values.append(getattr(instance, name))
            placeholders.append("?")

        fields_str = ", ".join(field_names)
        placeholders_str = ", ".join(placeholders)

        self.execute(
            f"""
INSERT INTO {table_name}
    ({fields_str})
VALUES
    ({placeholders_str})
""",
            values,
        )

    def all(
        self,
        cls: Type[T],
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        order_by: Optional[str] = None,
    ) -> List[T]:
        table_name = _make_table_name(cls.__name__)
        if limit is not None:
            assert isinstance(limit, int)
        if offset is not None:
            assert isinstance(offset, int)

        sql = f"SELECT * FROM {table_name}"

        if order_by:
            # Careful: users cannot provide this as it's susceptible to
            # sql injection.
            sql += f" ORDER BY {order_by}"

        if limit:
            sql += f" LIMIT {limit}"

        if offset:
            sql += f" OFFSET {offset}"

        return self.select(cls, sql)

    def select(self, cls: Type[T], sql: str, values: Optional[list] = None):
        with self.cursor() as cursor:
            self.execute_query(cursor, sql, values)
            return [cls(*x) for x in cursor.fetchall()]

    def first(
        self,
        cls: Type[T],
        query: Optional[str] = None,
        values: Optional[List[Any]] = None,
    ) -> T:
        """
        Note: we want to use plain sql, not a bunch of ORM to query.

        Maybe we can use the same structure as:
        https://github.com/kruxia/sqly

        to build the SQL though (but right now the query is just
        the actual SQL with placeholders and the values should be
        passed separately).

        Example:

        db.first(
                ActionPackage,
                "SELECT * FROM action_package WHERE id = ?",
                [action.action_package_id],
            )

        Raises:
            KeyError if no entries were returned in the query.
        """
        if not query:
            table_name = _make_table_name(cls.__name__)
            query = f"SELECT * FROM {table_name}"

        with self.cursor() as cursor:
            self.execute_query(cursor, query, values=values)
            one = cursor.fetchone()
            if one is None:
                raise KeyError("Query returned no entries.")
            return cls(*one)

    def list_table_names(self) -> List[str]:
        with self.cursor() as cursor:
            self.execute_query(
                cursor,
                """
SELECT
    name
FROM
    sqlite_schema
WHERE
    type ='table' AND
    name NOT LIKE 'sqlite_%';
""",
            )
            return [x[0] for x in cursor.fetchall()]

    def list_table_and_columns(self) -> Dict[str, List[str]]:
        with self.cursor() as cursor:
            self.execute_query(
                cursor,
                """
SELECT m.name as tableName, 
       p.name as columnName
FROM sqlite_master m
left outer join pragma_table_info((m.name)) p
     on m.name <> p.name
order by tableName, columnName;
""",
            )
            found: Dict[str, List[str]] = {}
            for table_name, column_name in cursor.fetchall():
                columns = found.get(table_name)
                if not columns:
                    columns = found[table_name] = []
                columns.append(column_name)
            return found

    def list_indexes(self) -> List[List[str]]:
        with self.cursor() as cursor:
            self.execute_query(
                cursor,
                """
SELECT 
    m.tbl_name as table_name,
    il.name as index_name,
    ii.name as column_name,
    CASE il.origin when 'pk' then 1 else 0 END as is_primary_key,
    CASE il.[unique] when 1 then 0 else 1 END as non_unique,
    il.[unique] as is_unique,
    il.partial,
    il.seq as sequence_in_index,
    ii.seqno as sequence_in_column
FROM sqlite_master AS m,
    pragma_index_list(m.name) AS il,
    pragma_index_info(il.name) AS ii
WHERE 
    m.type = 'table'
    and m.tbl_name = 'YOUR TABLENAME HERE'
GROUP BY
    m.tbl_name,
    il.name,
    ii.name,
    il.origin,
    il.partial,
    il.seq
ORDER BY index_name,il.seq,ii.seqno""",
            )
            return [x for x in cursor.fetchall()]

    def list_whole_db(self) -> Dict[str, List[Dict[str, Any]]]:
        table_to_contents: Dict[str, List[Dict[str, Any]]] = {}
        for table, cls in self._table_name_to_cls.items():
            with self.cursor() as cursor:
                self.execute_query(cursor, f"SELECT * FROM {table}")

                cls = self._table_name_to_cls[table]
                field_names = list(self._iter_name_fields(cls))
                rows = []
                for row in cursor.fetchall():
                    rows.append(dict(itertools.zip_longest(field_names, row)))

                table_to_contents[table] = rows
        return table_to_contents

    def load_whole_db(self, contents: Dict[str, List[Dict[str, Any]]]) -> None:
        with self.transaction():
            for table, table_rows in contents.items():
                cls = self._table_name_to_cls[table]
                for row in table_rows:
                    instance = cls(**row)
                    self.insert(instance)

    def _print_sql(self, sql: str, values: Optional[list] = None):
        func_name = sys._getframe(1).f_code.co_name
        print(f"db.{func_name}('''\n{sql.strip()}\n''', {values!r})\n")

    def _raise_execute_error(self, msg):
        if self.verbose:
            print(msg, file=sys.stderr)
        raise DBError(msg)

    def execute_query(
        self, cursor: sqlite3.Cursor, sql: str, values: Optional[list] = None
    ):
        """
        Executes a query which will NOT change the database (and should return values).

        No write-lock needed.
        """
        if self.verbose:
            self._print_sql(sql, values)

        try:
            if values:
                cursor.execute(sql, values)
            else:
                cursor.execute(sql)
        except Exception:
            self._raise_execute_error(
                f"Error running sql query: {sql!r} with values: {values!r}"
            )

    def execute_update_returning(
        self, cursor: sqlite3.Cursor, sql: str, values: Optional[list] = None
    ):
        """
        Executes a query which will NOT change the database (and should return values).

        No write-lock needed.
        """
        if self.verbose:
            self._print_sql(sql, values)
        try:
            if not self.in_transaction():
                raise DBError(
                    "When running an sql that changes the DB, it's expected that "
                    "a transaction is in place."
                )
            with self._write_lock:
                if values:
                    cursor.execute(sql, values)
                else:
                    cursor.execute(sql)
        except Exception:
            self._raise_execute_error(
                f"Error running sql: {sql!r} with values: {values!r}"
            )

    def execute(self, sql: str, values: Optional[list] = None) -> None:
        """
        Executes a query which will change the database.

        Requires the write-lock to be acquired since SQLite can't deal with
        writes in multiple threads concurrently.
        """

        if self.verbose:
            self._print_sql(sql, values)
        try:
            if not self.in_transaction():
                raise DBError(
                    "When running an sql that changes the DB, it's expected that "
                    "a transaction is in place."
                )
            conn = self._tlocal.conn
            assert conn is not None
            with self._write_lock:
                if values:
                    conn.execute(sql, values)
                else:
                    conn.execute(sql)
        except Exception:
            self._raise_execute_error(
                f"Error running sql: {sql!r} with values: {values!r}"
            )

    def register_classes(self, classes: List[Type]) -> None:
        if self._table_name_to_cls:
            values = set(self._table_name_to_cls.values())
            if values != set(classes):
                raise RuntimeError(
                    "The classes were already registered "
                    "(and do not match the new values)."
                )
            # i.e.: they were already registered.
            return

        self._classes = classes
        for cls in classes:
            self._table_name_to_cls[_make_table_name(cls.__name__)] = cls

    def initialize(self, classes: List[Type]) -> None:
        """
        Initializes the internal structure of the tables as needed
        (but doesn't really create the db -- use 'create_tables'
        if needed).
        """
        self.register_classes(classes)

    def create_tables(self, db_rules: Optional[DBRules] = None):
        if db_rules is None:
            db_rules = DBRules()

        sqls = []
        for cls in self._classes:
            sql = self.create_table_sql(cls, db_rules)
            sqls.append(sql)

            sqls.extend(self.create_indexes_sql(cls, db_rules))

        with self.connect():
            with self.transaction():
                for sql in sqls:
                    self.execute(sql)

    def create_indexes_sql(self, cls: Type, db_rules: DBRules) -> List[str]:
        table_name = _make_table_name(cls.__name__)

        columns: List[str] = []
        for name in self._iter_name_fields(cls):
            field_full_name = f"{cls.__name__}.{name}"
            if field_full_name in db_rules.unique_indexes:
                columns.append(name)

        ret: List[str] = []
        for column in columns:
            sql = f"""
CREATE UNIQUE INDEX {table_name}_{column}_index ON {table_name}({column});
"""
            ret.append(sql)
        return ret

    def create_table_sql(self, cls: Type, db_rules: DBRules) -> str:
        table_name = _make_table_name(cls.__name__)

        fields = []
        foreign_keys = []
        for name, field_cls in self._iter_name_and_name_cls_fields(cls):
            field_full_name = f"{cls.__name__}.{name}"
            fields.append(
                self._get_field_create_sql(
                    cls, field_full_name, name, field_cls, db_rules
                )
            )

            if field_full_name in db_rules.foreign_keys:
                assert name.endswith("_id")

                foreign_table = name[:-3]
                if foreign_table not in self._table_name_to_cls:
                    raise RuntimeError(
                        f"Error: unexpected foreign reference: {foreign_table} "
                        f"(for field: {name})"
                    )
                foreign_keys.append(
                    f"FOREIGN KEY ({name}) " f"REFERENCES {foreign_table}(id)"
                )

        fields.extend(foreign_keys)
        fields_str = ",\n    ".join(fields)
        sql = f"""
CREATE TABLE IF NOT EXISTS {table_name}(
    {fields_str}  
)
        """
        return sql

    def _get_field_create_sql(
        self,
        cls,
        field_full_name: str,
        name: str,
        field_cls: Type,
        db_rules: DBRules,
    ) -> str:
        primary_key = name == "id"

        if field_cls.__name__ == "Optional":
            not_none = [x for x in field_cls.__args__ if x != NoneType]
            assert len(not_none) == 1, f"Expected one not none in: {field_cls.__args__}"
            field_cls = not_none[0]
            not_null = False
        else:
            not_null = True

        if field_cls == int:
            use = "INTEGER"

        elif field_cls == str:
            use = "TEXT"

        elif field_cls == bool and not not_null:
            use = "INTEGER"

        elif field_cls == bool and not_null:
            default_value = getattr(cls, name)
            if default_value:
                use = f"INTEGER CHECK({name} IN (0, 1)) NOT NULL DEFAULT 1"
            else:
                use = f"INTEGER CHECK({name} IN (0, 1)) NOT NULL DEFAULT 0"

        elif field_cls == datetime.datetime:
            raise RuntimeError(
                f"Datetime not supported (field: {name}). Please "
                "use str and use utility functions to convert back and forth."
            )

        elif field_cls == float:
            use = "REAL"

        else:
            raise RuntimeError(f"Unsupported type: {field_cls}")

        use = f"{name} {use}"

        if not_null and field_cls != bool:
            use = f"{use} NOT NULL"

        if primary_key:
            use = f"{use} PRIMARY KEY"

        return f"{use}"
